#  _Exponential.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, Tuple

from ....core.matrix.factory import filled_like
from ....core.matrix import Matrix
from ._NegEntropyApproximationFunction import NegEntropyApproximationFunction


class Exponential(NegEntropyApproximationFunction):
    """
    Exponential Negative Entropy Approximation Function.
    """
    def apply(self, x: Optional[Matrix]) -> Tuple[Matrix, Matrix]:
        x_pow_2: Matrix = x.pow_elementwise(2)
        exp: Matrix = x_pow_2.div(2).mul(-1).exp()

        gx: Matrix = x.mul_elementwise(exp)

        ones: Matrix = filled_like(gx, 1)
        g_x: Matrix = ones.sub(x_pow_2).mul_elementwise(exp).mean(1)

        return gx, g_x



#  _DIPLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from enum import Enum, auto
from typing import Optional, List

from ._AbstractSingleResponsePLS import AbstractSingleResponsePLS
from ...core import ZERO, real, ONE, NAN
from ...core.matrix import Matrix, factory
from ...meta import switch, break_, case
from ...transformation import Center


class DIPLS(AbstractSingleResponsePLS):
    def __init__(self):
        super().__init__()
        self.model_adaption_strategy: Optional[ModelAdaptionStrategy] = None  # Model adaption strategy
        self.ns: int = 0  # Number of source train samples
        self.nt: int = 0  # Number of target train samples
        self.lambda_: real = ONE  # Lambda parameter
        self.b_0: real = ZERO  # Response mean
        self.T: Optional[Matrix] = None  # Loadings
        self.T_s: Optional[Matrix] = None  # Source domain loadings
        self.T_t: Optional[Matrix] = None  # Target domain loadings
        self.P: Optional[Matrix] = None  # Scores
        self.P_s: Optional[Matrix] = None  # Source domain scores
        self.P_t: Optional[Matrix] = None  # Target domain scores
        self.W_di: Optional[Matrix] = None  # Weights
        self.b_di: Optional[Matrix] = None  # Regression coefficients
        self.X_center: Optional[Center] = None  # X center
        self.X_s_center: Optional[Center] = None  # Source domain center
        self.X_t_center: Optional[Center] = None  # Target domain center

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.model_adaption_strategy = ModelAdaptionStrategy.UNSUPERVISED
            self.lambda_ = ONE
            self.X_center = Center()
            self.X_t_center = Center()
            self.X_s_center = Center()
        else:
            return super().initialize(predictors, response)

    def reset(self):
        super().reset()
        self.T = None
        self.T_s = None
        self.T_t = None
        self.P = None
        self.P_s = None
        self.P_t = None
        self.W_di = None
        self.b_di = None
        self.b_0 = NAN

    @staticmethod
    def validate_lambda_(value: real) -> bool:
        return abs(value) >= 1e-8

    def get_matrix_names(self) -> List[str]:
        return ['T', 'Ts', 'Tt',
                'Wdi',
                'P', 'Ps', 'Pt',
                'bdi']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        map = {'T': self.T,
               'Ts': self.T_s,
               'Tt': self.T_t,
               'P': self.P,
               'Ps': self.P_s,
               'Pt': self.P_t,
               'Wdi': self.W_di,
               'bdi': self.b_di}

        if name in map:
            return map[name]

        return None

    def has_loadings(self) -> bool:
        return True

    def get_loadings(self) -> Optional[Matrix]:
        return self.T

    def can_predict(self) -> bool:
        return True

    def do_transform(self, predictors: Matrix) -> Matrix:
        return self.X_center.transform(predictors).mul(self.W_di)

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        num_features: int = predictors.num_columns()
        I: Matrix = factory.eye(num_features)
        c: Optional[Matrix] = None

        # Check if correct initialization method was called
        if self.ns == 0 or self.nt == 0:
            return 'DIPLS must be initialized with one of the three following methods:\n' +\
                   ' - initializeSupervised' +\
                   ' - initializeSemiSupervised' +\
                   ' - initializeUnsupervisedSupervised'

        # Check if sufficient source and target samples exist
        if self.ns == 1 or self.nt == 1:
            return 'Number of source and target samples has to be > 1.'

        # Initialise Xs, Xt, X, y
        with switch(self.model_adaption_strategy):
            if case(ModelAdaptionStrategy.UNSUPERVISED):
                X_s: Matrix = predictors.get_rows((0, self.ns))
                X_t: Matrix = predictors.get_rows((self.ns, predictors.num_rows()))

                X: Matrix = X_s.copy()
                y = response
                break_()
            if case(ModelAdaptionStrategy.SUPERVISED):
                X_s: Matrix = predictors.get_rows((0, self.ns))
                X_t: Matrix = predictors.get_rows((self.ns, predictors.num_rows()))

                X: Matrix = predictors
                y = response
                break_()
            if case(ModelAdaptionStrategy.SEMISUPERVISED):
                X_s: Matrix = predictors.get_rows((0, self.ns))
                X_t: Matrix = predictors.get_rows((self.ns, predictors.num_rows()))

                X: Matrix = predictors.get_rows((0, 2 * self.ns))
                y = response
                break_()

        # Center X, Xs, Xt
        X = self.X_center.transform(X)
        X_s = self.X_s_center.transform(X_s)
        X_t = self.X_t_center.transform(X_t)

        # Center y
        self.b_0 = y.mean(-1).as_real()
        y = y.sub(self.b_0)

        # Start loop over number of components
        for a in range(self.num_components):

            # Calculate domain invariant weights
            y_norm2_squared: real = y.norm2_squared()
            w_di_LHS: Matrix = y.transpose().mul(X).div(y_norm2_squared)
            X_s_t_X_s: Matrix = X_s.transpose().mul(X_s).mul(ONE / (self.ns - ONE))
            X_t_t_X_t: Matrix = X_t.transpose().mul(X_t).mul(ONE / (self.nt - ONE))
            X_s_diff_X_t: Matrix = X_s_t_X_s.sub(X_t_t_X_t)
            w_di_RHS: Matrix = I.add(X_s_diff_X_t.mul(self.lambda_ / (2 * y_norm2_squared))).inverse()
            w_di: Matrix = w_di_LHS.mul(w_di_RHS).transpose()
            w_di = w_di.normalized()

            # Calculate loadings
            t: Matrix = X.mul(w_di)
            t_s: Matrix = X_s.mul(w_di)
            t_t: Matrix = X_t.mul(w_di)

            # Calculate scores
            p: Matrix = (t.transpose().mul(t)).inverse().mul(t.transpose()).mul(X)
            p_s: Matrix = (t_s.transpose().mul(t_s)).inverse().mul(t_s.transpose()).mul(X_s)
            p_t: Matrix = (t_t.transpose().mul(t_t)).inverse().mul(t_t.transpose()).mul(X_t)
            ca: Matrix = (t.transpose().mul(t)).inverse().mul(y.transpose()).mul(t)

            # Deflate X, Xs, Xt, y
            X = X.sub(t.mul(p))
            X_s = X_s.sub(t_s.mul(p_s))
            X_t = X_t.sub(t_t.mul(p_t))
            y = y.sub(t.mul(ca))

            # Collect
            c = self.concat(c, ca)

            self.T = self.concat(self.T, t)
            self.T_s = self.concat(self.T_s, t_s)
            self.T_t = self.concat(self.T_t, t_t)

            self.P = self.concat(self.P, p.transpose())
            self.P_s = self.concat(self.P_s, p_s.transpose())
            self.P_t = self.concat(self.P_t, p_t.transpose())

            self.W_di = self.concat(self.W_di, w_di)

        # Calculate regression coefficients
        self.b_di = self.W_di.mul((self.P.transpose().mul(self.W_di)).inverse()).mul(c.transpose())

        return None

    def concat(self, A: Optional[Matrix], a: Matrix) -> Matrix:
        """
        Concat A along columns with a. If A is None, return a.

        :param A:   Base matrix.
        :param a:   Column vector.
        :return:    Concatenation of A and a.
        """
        if A is None:
            return a
        else:
            return A.concat_along_columns(a)

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        recentered: Optional[Matrix] = None

        # Recenter
        with switch(self.model_adaption_strategy):
            if case(ModelAdaptionStrategy.UNSUPERVISED):
                recentered = self.X_t_center.transform(predictors)
                break_()
            if case(ModelAdaptionStrategy.SUPERVISED,
                    ModelAdaptionStrategy.SEMISUPERVISED):
                recentered = self.X_center.transform(predictors)
                break_()

        # Predict
        regression: Matrix = recentered.mul(self.b_di)

        # Add response means
        result: Matrix = regression.add(self.b_0)
        return result

    def initialize_unsupervised(self,
                                predictors_source_domain: Matrix,
                                predictors_target_domain: Matrix,
                                response_source_domain: Matrix) -> str:
        """
        Unsupervised initialisation.

        :param predictors_source_domain:    Predictors from source domain.
        :param predictors_target_domain:    Predictors from target domain.
        :param response_source_domain:      Response from source domain.
        :return:                            Result, None if no errors, else error string.
        """
        self.ns = predictors_source_domain.num_rows()
        self.nt = predictors_target_domain.num_rows()
        self.model_adaption_strategy = ModelAdaptionStrategy.UNSUPERVISED
        X: Matrix = predictors_source_domain.concat_along_rows(predictors_target_domain)
        y: Matrix = response_source_domain

        return self.initialize(X, y)

    def initialize_supervised(self,
                              predictors_source_domain: Matrix,
                              predictors_target_domain: Matrix,
                              response_source_domain: Matrix,
                              response_target_domain: Matrix) -> str:
        """
        Supervised initialisation.

        :param predictors_source_domain:    Predictors from source domain.
        :param predictors_target_domain:    Predictors from target domain.
        :param response_source_domain:      Response from source domain.
        :param response_target_domain:      Response from target domain.
        :return:                            Result, None if no errors, else error string.
        """
        self.ns = predictors_source_domain.num_rows()
        self.nt = predictors_target_domain.num_rows()
        self.model_adaption_strategy = ModelAdaptionStrategy.SUPERVISED
        X: Matrix = predictors_source_domain.concat_along_rows(predictors_target_domain)
        y: Matrix = response_source_domain.concat_along_rows(response_target_domain)

        return self.initialize(X, y)

    def initialize_semisupervised(self,
                                  predictors_source_domain: Matrix,
                                  predictors_target_domain: Matrix,
                                  predictors_target_domain_unlabeled: Matrix,
                                  response_source_domain: Matrix,
                                  response_target_domain: Matrix) -> str:
        """
        Semisupervised initialisation.

        :param predictors_source_domain:            Predictors from source domain.
        :param predictors_target_domain:            Predictors from target domain.
        :param predictors_target_domain_unlabeled:  Predictors from target domain without labels.
        :param response_source_domain:              Response from source domain.
        :param response_target_domain:              Response from target domain.
        :return:                                    Result, None if no errors, else error string.
        """
        self.ns = predictors_source_domain.num_rows()
        self.nt = predictors_target_domain.num_rows()
        self.model_adaption_strategy = ModelAdaptionStrategy.SEMISUPERVISED
        X: Matrix = predictors_source_domain\
                    .concat_along_rows(predictors_target_domain)\
                    .concat_along_rows(predictors_target_domain_unlabeled)
        y: Matrix = response_source_domain.concat_along_rows(response_target_domain)

        return self.initialize(X, y)


class ModelAdaptionStrategy(Enum):
    UNSUPERVISED = auto()
    SUPERVISED = auto()
    SEMISUPERVISED = auto()



#  _StandardizeTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..test.misc import Test
from ..transformation import AbstractTransformationTest
from wai.ma.core import real
from wai.ma.core.matrix import Matrix, helper
from wai.ma.transformation import Standardize


class StandardizeTest(AbstractTransformationTest[Standardize]):
    @Test
    def check_zero_variance(self):
        X: Matrix = self.input_data[0]
        transform: Matrix = self.subject.transform(X)

        actual_mean: real = transform.mean()
        expected_mean: real = real(0.0)
        expected_std: real = real(1.0)

        self.assertAlmostEqual(expected_mean, actual_mean, delta=1e-7)

        for i in range(transform.num_columns()):
            actual_std: real = helper.stdev(transform, i)
            self.assertAlmostEqual(expected_std, actual_std, delta=1e-7)

    def instantiate_subject(self) -> Standardize:
        return Standardize()


#  _SavitzkyGolay2.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ._SavitzkyGolay import SavitzkyGolay


class SavitzkyGolay2(SavitzkyGolay):
    def __init__(self):
        super().__init__()
        self.num_points: int = 3

        # Delete num_points_left/right
        del self.num_points_left
        del self.num_points_right

    def __getattribute__(self, item):
        # Alias num_points_left/right to num_points
        if item in {'num_points_left', 'num_points_right'}:
            return self.num_points

        return super().__getattribute__(item)

    @staticmethod
    def validate_num_points(value: int):
        if value < 0:
            raise ValueError('num_points must be at least 0')



#  _EPOTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from ._GLSWTest import GLSWTest
from ...test.misc import TestRegression, TestDataset
from wai.ma.algorithm.glsw import EPO


class EPOTest(GLSWTest[EPO]):
    @TestRegression
    def n_1(self):
        self.subject.N = 1

    @TestRegression
    def n_3(self):
        self.subject.N = 3

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS]

    def instantiate_subject(self) -> EPO:
        return EPO()


#  _DIPLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from ._AbstractPLSTest import AbstractPLSTest
from ...test.misc import TestRegression, Tags, Test
from wai.ma.algorithm.pls import DIPLS
from wai.ma.core.matrix import Matrix
from wai.ma.core.matrix.factory import randn_like


class DIPLSTest(AbstractPLSTest[DIPLS]):
    @TestRegression
    def lambda_01(self):
        self.subject.lambda_ = 0.01

    def setup_regressions(self, subject: DIPLS, input_data: List[Matrix]):
        x_source_domain: Matrix = input_data[0]
        y_source_domain: Matrix = input_data[1]
        x_target_domain: Matrix = x_source_domain.add(randn_like(x_source_domain, 0, 1, 2))
        y_target_domain: Matrix = y_source_domain.add(randn_like(y_source_domain, 1, 1, 2))
        x_target_domain_unlabeled: Matrix = x_source_domain.add(randn_like(x_source_domain, 100, 1, 2))

        # Initialise supervised
        subject.initialize_supervised(x_source_domain, x_target_domain, y_source_domain, y_target_domain)
        self.add_default_pls_matrices(subject, x_target_domain, Tags.SUPERVISED)
        subject.reset()

        # Initialise unsupervised
        subject.initialize_unsupervised(x_source_domain, x_target_domain, y_source_domain)
        self.add_default_pls_matrices(subject, x_target_domain, Tags.UNSUPERVISED)
        subject.reset()

        # Initialise semisupervised
        subject.initialize_semisupervised(x_source_domain, x_target_domain, x_target_domain_unlabeled,
                                          y_source_domain, y_target_domain)
        self.add_default_pls_matrices(subject, x_target_domain, Tags.SEMISUPERVISED)

    @Test
    def check_transformed_num_components(self):
        X: Matrix = self.input_data[0]
        X_2: Matrix = X.add(randn_like(X, 0, 0, 2))
        Y: Matrix = self.input_data[1]

        for i in range(1, 5):
            self.subject.num_components = i
            self.subject.initialize_unsupervised(X, X_2, Y)
            transform: Matrix = self.subject.transform(X)
            self.assertEqual(i, transform.num_columns())

            # Reset
            self.subject = self.instantiate_subject()

    def instantiate_subject(self) -> DIPLS:
        return DIPLS()



#  _LinearKernel.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ...core import real
from ...core.matrix import Matrix
from ._AbstractKernel import AbstractKernel


class LinearKernel(AbstractKernel):
    """
    Linear Kernel.

    K(x_i,y_j)=x_i^T*y_j
    or also
    K(X, Y)=X*Y^T
    """
    def apply_vector(self, x: Matrix, y: Matrix) -> real:
        return x.vector_dot(y)

    def apply_matrix(self, X: Matrix, Y: Optional[Matrix] = None) -> Matrix:
        if Y is None:
            Y = X
        return X.mul(Y.transpose())

    def __str__(self):
        return 'Linear Kernel: K(x,y)=x^T*y'



#  _Tags.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Tags that should be used when calling AbstractRegressionTest.add_regression(str, Union[Matrix, real]).
"""

TRANSFORM = 'transform'
INVERSE_TRANSFORM = 'inverse-transform'
LOADINGS = 'loadings'
SCORES = 'scores'
PREDICTIONS = 'predictions'
PROJECTION = 'projection'
MATRIX = 'matrix'
SUPERVISED = 'supervised'
SEMISUPERVISED = 'semisupervised'
UNSUPERVISED = 'unsupervised'



#  _GLSW.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ...core import SupervisedFilter
from .._AbstractAlgorithm import AbstractAlgorithm
from ...core.matrix import Matrix, factory
from ...transformation import Center


class GLSW(AbstractAlgorithm, SupervisedFilter):
    def __init__(self):
        super().__init__()
        self.alpha: float = 1e-3
        self.G: Optional[Matrix] = None

    @staticmethod
    def validate_alpha(value: float) -> bool:
        return value > 0

    def reset(self):
        super().reset()
        self.G = None

    def initialize(self, x1: Optional[Matrix] = None, x2: Optional[Matrix] = None) -> Optional[str]:
        if x1 is None and x2 is None:
            super().initialize()
            self.alpha = 1e-3
        else:
            # Always work on copies
            x1 = x1.copy()
            x2 = x2.copy()

            self.reset()

            result: Optional[str] = self.check(x1, x2)

            if result is None:
                result = self.do_initialize(x1, x2)
                self.initialised = result is None

            return result

    def to_string(self) -> str:
        if self.initialised:
            return 'Generalized Least Squares Weighting. Projection Matrix shape: ' + self.G.shape_string()
        else:
            return 'Generalized Least Squares Weighting. Model not yet initialized.'

    def do_initialize(self, x1: Matrix, x2: Matrix) -> Optional[str]:
        super().initialize()

        C: Matrix = self.get_covariance_matrix(x1, x2)

        # SVD
        V: Matrix = self.get_eigenvector_matrix(C)
        D: Matrix = self.get_weight_matrix(C)

        # Projection matrix
        self.G = V.mul(D.inverse()).mul(V.transpose())

        return None

    def get_eigenvector_matrix(self, C: Matrix) -> Matrix:
        return C.get_eigenvalue_decomposition_V()

    def get_weight_matrix(self, C: Matrix) -> Matrix:
        # Get eigenvalues
        S_squared: Matrix = C.svd_S().pow_elementwise(2)

        # Weights
        D: Matrix = S_squared.div(self.alpha)
        D = D.add(factory.eye_like(D))
        D = D.sqrt()
        return D

    def get_covariance_matrix(self, x1: Matrix, x2: Matrix) -> Matrix:
        # Center X1, X2
        c1: Center = Center()
        c2: Center = Center()
        x1_centered: Matrix = c1.transform(x1)
        x2_centered: Matrix = c2.transform(x2)

        # Build difference
        X_d: Matrix = x2_centered.sub(x1_centered)

        # Covariance Matrix
        return X_d.transpose().mul(X_d)

    def do_transform(self, predictors: Matrix) -> Matrix:
        return predictors.mul(self.G)

    def transform(self, predictors: Matrix) -> Matrix:
        if not self.is_initialised():
            raise RuntimeError("Algorithm hasn't been initialized!")

        return self.do_transform(predictors)

    def check(self, x1: Matrix, x2: Matrix) -> Optional[str]:
        if x1 is None:
            return 'No x1 matrix provided!'
        if x2 is None:
            return 'No x2 matrix provided!'
        if not x1.same_shape_as(x2):
            return 'Matrices x1 and x2 must have the same shape'
        return None


#  _AbstractRegressionTestMeta.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import GenericMeta, Tuple, List, Dict, Any
from unittest import TestCase, defaultTestLoader

from .misc import TEST_TAG
from wai.ma.meta import has_tag, will_be_abstract


class AbstractRegressionTestMeta(GenericMeta):
    """
    Meta-class which adds the TestCase class as a base of any classes
    inheriting from AbstractRegressionTest. This is because making
    AbstractRegressionTest inherit from TestCase in the normal fashion
    causes unittest to try to instantiate it, which fails because it
    is abstract. An alternative would be to override the
    __subclasscheck__ method of the meta-class, but for some reason
    it wasn't being called (possible short-cutting in python implementation).
    """
    def __new__(self, name, bases, namespace, **kwargs):
        bases = self.augment_bases(bases, namespace)
        self.ensure_tests(namespace)
        return super().__new__(self, name, bases, namespace, **kwargs)

    @staticmethod
    def augment_bases(bases: Tuple, namespace: Dict) -> Tuple:
        """
        Augments the base classes for the new class with the TestCase class.

        :param bases:       The proposed bases for the class.
        :param namespace:   The namespace of the class.
        :return:            The base classes augmented with the TestCase class.
        """
        if not will_be_abstract(bases, namespace):
            if TestCase not in bases:
                return (*bases, TestCase)
            else:
                return bases
        else:
            if TestCase in bases:
                return tuple(base for base in bases if base is not TestCase)
            else:
                return bases

    @staticmethod
    def ensure_tests(namespace: Dict[str, Any]):
        """
        Makes sure that all methods marked with the test tag will be picked up by unittest.

        :param namespace:   The namespace of the new class.
        :return:            The namespace augmented with the unit-test methods.
        """
        prefix = defaultTestLoader.testMethodPrefix
        new_items = {}  # Required so namespace doesn't change during iteration
        for name, method in namespace.items():
            if callable(method) and has_tag(method, TEST_TAG):
                if not name.startswith(prefix):
                    test_name = prefix + '_' + name
                    if test_name in namespace:
                        continue  # Test method already present
                    new_items[test_name] = method
        namespace.update(new_items)



#  _MatrixTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from ...test.misc import TestDataset, Test, Tags, TestRegression
from ...test import AbstractRegressionTest
from wai.ma.core.matrix import Matrix, factory


class MatrixTest(AbstractRegressionTest[Matrix]):
    @TestRegression
    def transpose(self):
        self.subject = self.subject.transpose()

    @TestRegression
    def get_eigenvectors(self):
        self.subject = self.subject.get_eigenvectors()

    @TestRegression
    def get_eigenvalues(self):
        self.subject = self.subject.get_eigenvalues()

    @TestRegression
    def norm2(self):
        self.subject = self.input_data[0].norm2()

    @TestRegression
    def norm1(self):
        self.subject = self.input_data[0].norm1()

    @Test
    def column_is_sub_matrix(self):
        for i in range(self.subject.num_columns()):
            column = self.subject.get_column(i)
            sub_matrix = self.subject.get_sub_matrix((0, self.subject.num_rows()),
                                                     (i, i + 1))
            self.assertEqual(column, sub_matrix)

    def setup_regressions(self, subject: Matrix, input_data: List[Matrix]):
        # Only regression testing the final configuration of the matrix
        self.add_regression(Tags.MATRIX, subject)

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS]

    def instantiate_subject(self) -> Matrix:
        return factory.randn(7, 7, seed=2)




#  _PRM.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, List


from ._AbstractSingleResponsePLS import AbstractSingleResponsePLS
from ._SIMPLS import SIMPLS
from ...core import real, ZERO, ONE
from ...core.matrix import Matrix, factory
from ...core.utils import sqrt
from ...meta import switch, case


class PRM(AbstractSingleResponsePLS):
    """
    Partial robust M-regression as described in
    <a href="https://www.sciencedirect.com/science/article/abs/pii/S0169743905000638">
        Partial robust M-regression
    </a>

    Parameters:
    - c: Tuning constant for Fair weight function. Higher values result in a
         flatter function. c=Infinity is equal to SIMPLS result.
    - tol: Iterative convergence tolerance
    - maxIter: Maximum number of iterations
    - numSimplsCoefficients: Number of SIMPLS coefficients
    """
    def __init__(self):
        super().__init__()
        self.tol: real = ZERO  # Loop improvement tolerance
        self.max_iter: int = 0  # Loop maximum number of iterations
        self.C: real = ONE  # C parameter
        self.W_r: Optional[Matrix] = None  # Residual weights N x 1
        self.W_x: Optional[Matrix] = None  # Leverage weights N x 1
        self.T: Optional[Matrix] = None  # T score matrix from SIMPLS algorithm
        self.gamma: Optional[Matrix] = None  # Gamma regression coefficients from SIMPLS algorithm
        self.num_SIMPLS_coefficients: int = 0  # The number of SIMPLS coefficients in W to keep (0 keep all)
        self.final_regression_coefficients: Optional[Matrix] = None  # Final regression coefficients
        self.simpls: Optional[SIMPLS] = None  # SIMPLS algorithm

    @staticmethod
    def validate_num_SIMPLS_coefficients(value: int) -> bool:
        return True

    @staticmethod
    def validate_max_iter(value: int) -> bool:
        return value >= 0

    @staticmethod
    def validate_tol(value: real) -> bool:
        return value >= 0

    @staticmethod
    def validate_C(value: real) -> bool:
        return abs(value) >= 1e-10

    def reset(self):
        super().reset()
        self.W_r = None
        self.W_x = None
        self.final_regression_coefficients = None
        self.gamma = None
        self.T = None
        self.simpls = None

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.C = real(4)
            self.tol = real(1e-6)
            self.max_iter = 500
            self.num_SIMPLS_coefficients = -1
        else:
            return super().initialize(predictors, response)

    def get_matrix_names(self) -> List[str]:
        return ['B',
                'Wr',
                'Wx',
                'W']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        with switch(name):
            if case('B'):
                return self.final_regression_coefficients
            if case('Wr'):
                return self.W_r
            if case('Wx'):
                return self.W_x
            if case('W'):
                return self.W_r.mul_elementwise(self.W_x)
        return None

    def has_loadings(self) -> bool:
        return False

    def get_loadings(self) -> Optional[Matrix]:
        return None

    def to_string(self) -> str:
        return ''

    def fair_function(self, z: real, c: real) -> real:
        """
        Fair function implementation.

        f(z,c) = 1/(1+|z/c|)^2

        :param z:   First parameter.
        :param c:   Second parameter.
        :return:    Fair function result.
        """
        return real(1.0 / pow(1.0 + abs(z / c), 2))

    def init_weights(self, X: Matrix, y: Matrix):
        """
        Initialise the residual and leverage weights.

        :param X:   Predictor matrix.
        :param y:   Response matrix.
        """
        self.update_residual_weights(X, y)
        self.update_leverage_weights(X)

    def update_leverage_weights(self, T: Matrix):
        """
        Update the leverage weights based on the score matrix T.

        :param T:   Score matrix.
        """
        n: int = T.num_rows()
        self.W_x = factory.zeros(n, 1)

        row_L1_median: Matrix = self.geometric_median(T)
        distances_to_median: Matrix = self.c_dist(T, row_L1_median)

        median_of_dists_to_median: real = distances_to_median.median()

        # Calculate w_xi by f(zi, c) with zi = (distance_i to median) / (median of distances to median)
        for i in range(n):
            dist_to_median: real = distances_to_median.get(i, 0)
            w_xi: real = self.fair_function(real(dist_to_median / median_of_dists_to_median), self.C)
            self.W_x.set(i, 0, w_xi)

    def update_residual_weights(self, X: Matrix, y: Matrix):
        """
        Update the residual weights based on the new residuals.

        :param X:   Predictors.
        :param y:   Response.
        """
        n: int = X.num_rows()
        self.W_r = factory.zeros(n, 1)

        residuals: Matrix = factory.zeros(n, 1)

        # Check if this is the first iteration
        is_first_iteration: bool = self.T is None and self.gamma is None

        y_i_hat: real = y.median()
        for i in range(n):
            # Use t_i * gamma as estimation if iteration > 0
            if not is_first_iteration:
                y_i_hat = self.T.get_row(i).mul(self.gamma).as_real()

            # Calculate residual
            y_i: real = y.get(i, 0)
            r_i: real = real(y_i - y_i_hat)
            residuals.set(i, 0, r_i)

        # Get estimate of residual scale
        sigma: real = self.median_absolute_deviation(residuals)
        residuals = residuals.div(sigma)

        # Calculate weights
        for i in range(n):
            r_i: real = residuals.get(i, 0)
            w_ri: real = self.fair_function(r_i, self.C)
            self.W_r.set(i, 0, w_ri)

    def median_absolute_deviation(self, v: Matrix) -> real:
        """
        Mean Absolute Deviation:
        MAD(v) = median_i | v_i - median_j v_j |

        :param v:   Input vector.
        :return:    MAD result.
        """
        return v.sub(v.median()).abs().median()

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Trains using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        X: Matrix = predictors
        y: Matrix = response
        U: Optional[Matrix] = None

        # If X: n x p and p > n, use SVD to replace X with n x n matrix
        # See also: Remark 2 in paper
        has_more_columns_than_rows: bool = X.num_columns() > X.num_rows()
        if has_more_columns_than_rows:
            X_t: Matrix = X.transpose()
            U = X_t.svd_U()
            V: Matrix = X_t.svd_V()
            S: Matrix = X_t.svd_S()

            # Replace X with n x n matrix
            X = V.mul(S)

        # 1) Compute robust starting values fro residual and leverage weights
        self.init_weights(X, y)

        gamma_old: Optional[Matrix] = None
        num_components: int = self.num_components
        self.gamma = factory.zeros(num_components, 1)
        iteration: int = 0

        # Loop until convergence of gamma
        while iteration == 0 or (self.gamma.sub(gamma_old).norm2_squared() < self.tol and iteration < self.max_iter):
            # 2) Perform PLS (SIMPLS) on reweighted data matrices
            X_p: Matrix = self.get_reweighted_matrix(X)
            y_p: Matrix = self.get_reweighted_matrix(y)

            self.simpls = SIMPLS()
            self.simpls.num_coefficients = self.num_SIMPLS_coefficients
            self.simpls.num_components = num_components
            self.simpls.initialize(X_p, y_p)

            # Get scores and regression coefficients
            gamma_old = self.gamma.copy()
            self.T = self.simpls.transform(X_p)
            self.gamma = self.simpls.get_matrix('Q').transpose()

            # Rescale t_i by 1/sqrt(w_i)
            for i in range(self.T.num_rows()):
                w_i_sqrt: real = sqrt(self.get_combined_weight(i))
                row_i_scaled: Matrix = self.T.get_row(i).div(w_i_sqrt)
                self.T.set_row(i, row_i_scaled)

            # Update weights
            self.update_residual_weights(X_p, y_p)
            self.update_leverage_weights(self.T)
            iteration += 1

        # Get the final regression coefficients from the latest SIMPLS run
        self.final_regression_coefficients = self.simpls.get_matrix('B')

        # If X has been replaced by US, the regression coefficients need to be
        # back-transformed into beta_hat = U*beta_p
        if has_more_columns_than_rows:
            self.final_regression_coefficients = U.mul(self.final_regression_coefficients)

        return None

    def can_predict(self) -> bool:
        return True

    def do_transform(self, predictors: Matrix) -> Matrix:
        return predictors.mul(self.simpls.get_matrix('W'))

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        return predictors.mul(self.final_regression_coefficients)

    def get_reweighted_matrix(self, A: Matrix) -> Matrix:
        return A.copy().scale_by_column_vector(self.W_r.mul_elementwise(self.W_x).sqrt())

    def get_combined_weight(self, i: int) -> real:
        w_xi: real = self.W_x.get_row(i).as_real()
        w_ri: real = self.W_r.get_row(i).as_real()
        return real(w_xi * w_ri)

    def geometric_median(self, X: Matrix) -> Matrix:
        """
        Geometric median according to
        <a href="https://en.wikipedia.org/wiki/Geometric_median">
            Geometric Median
        </a>

        Weiszfeld's algorithm.

        :param X:   Points.
        :return:    Geometric median of X.
        """
        # Initial guess
        guess: Matrix = X.mean(0)

        iteration: int = 0
        while iteration < self.max_iter:
            dists: Matrix = self.c_dist(X, guess)

            def _(value):
                if abs(value) < 1e-10:
                    return real(1.0 / 0.1)  # Fix zero distances
                else:
                    return real(1.0 / value)  # Invert
            dists = dists.apply_elementwise(_)

            nom: Matrix = X.scale_by_column_vector(dists).sum(0)
            denom: real = dists.sum(0).as_real()
            guess_next: Matrix = nom.div(denom)

            change: real = guess_next.sub(guess).norm2_squared()
            guess = guess_next
            if change < self.tol:
                break

            iteration += 1

        return guess

    def c_dist(self, X: Matrix, vector: Matrix) -> Matrix:
        """
        Distance function between all rows of X and a given row vector.

        :param X:       Input matrix with rows.
        :param vector:  Row vector to compare all rows of X to.
        :return:        Distances of each row r_i with the input vector.
        """
        dist: Matrix = factory.zeros(X.num_rows(), 1)
        for i in range(X.num_rows()):
            row_i: Matrix = X.get_row(i)
            d: real = row_i.sub(vector).norm2()
            dist.set(i, 0, d)
        return dist



#  _util.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import traceback
from typing import Tuple, Dict, Any


def print_stack_trace():
    """
    Prints the stack-trace to stderr.
    """
    traceback.print_exc()


def decorator_sequence(*decorators):
    """
    Helper method which creates a decorator that applies the given
    sub-decorators. Decorators are applied in reverse order given.

    @decorator_sequence(dec_1, dec_2, ...)
    def function(...):
        ...

    is equivalent to:

    @dec_1
    @dec_2
    ...
    def function(...):
        ...

    :param decorators:  The sub-decorators.
    :return:            A function which applies the given decorators.
    """
    def apply(obj):
        for dec in reversed(decorators):
            obj = dec(obj)
        return obj

    return apply


def get_abstract_methods(bases: Tuple, namespace: Dict[str, Any]):
    """
    Gets the names of the abstract methods that will result from
    a class created with the given base-class set and namespace.

    :param bases:       The base-class set.
    :param namespace:   The namespace.
    :return:            The set of abstract method names.
    """
    abstract_methods = set()

    for base in bases:
        abstract_methods.update(getattr(base, '__abstractmethods__', set()))

    for name, value in namespace.items():
        if is_abstract_function(value):
            abstract_methods.add(name)
        else:
            if name in abstract_methods:
                abstract_methods.remove(name)

    return abstract_methods


def will_be_abstract(bases: Tuple, namespace: Dict[str, Any]):
    """
    Determines if a class made with the given set of base-classes
    and namespace will be abstract or concrete.

    :param bases:       The set of base classes.
    :param namespace:   The namespace of the new class.
    :return:            True if the newly-created class will be abstract,
                        False if it will be concrete.
    """
    return len(get_abstract_methods(bases, namespace)) > 0


def is_abstract_function(func):
    return getattr(func, '__isabstractmethod__', False)



#  _InvalidAxisError.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


class InvalidAxisError(RuntimeError):
    def __init__(self, axis: int):
        super().__init__('Axis has to be either 0 or 1 but was ' + str(axis) + '.')



#  _PassThroughTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ._AbstractTransformationTest import AbstractTransformationTest
from ..test.misc import Test
from wai.ma.core.matrix import Matrix
from wai.ma.transformation import PassThrough


class PassThroughTest(AbstractTransformationTest[PassThrough]):
    @Test
    def result_unchanged(self):
        X: Matrix = self.input_data[0]
        transform: Matrix = self.subject.transform(X)

        is_equal: bool = X.sub(transform).abs().all(lambda v: v < 1e-7)
        self.assertTrue(is_equal)

    def instantiate_subject(self) -> PassThrough:
        return PassThrough()



#  _Tag.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


def Tag(tag: str):
    """
    Decorator that adds the given tag to the decorated function.

    :param tag:     The tag to add to the function.
    """
    def apply_tag(method):
        if not hasattr(method, '__tags'):
            method.__tags = set()
        method.__tags.add(tag)
        return method

    return apply_tag


def has_tag(method, tag: str) -> bool:
    """
    Checks if the given method has the given tag.

    :param method:  The method to check.
    :param tag:     The tag to check for.
    :return:        True if the tag exists on the method,
                    False if not.
    """
    return hasattr(method, '__tags') and tag in method.__tags


#  _AbstractRegressionTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from abc import abstractmethod
from typing import Optional, List, Union, TypeVar, Generic
from unittest import TestCase

from ._AbstractRegressionTestMeta import AbstractRegressionTestMeta
from .misc import TestDataset, TestRegression, REGRESSION_TAG
from wai.ma.core import real
from wai.ma.meta import print_stack_trace, has_tag
from ._RegressionManager import RegressionManager
from wai.ma.core.matrix import Matrix

T = TypeVar('T')


class AbstractRegressionTest(TestCase, Generic[T], metaclass=AbstractRegressionTestMeta):
    """
    Base class of all regression tests. Uses AbstractRegressionTestMeta to ensure
    TestCase is a base of all concrete sub-classes without actually being a base
    of this class. Hence method calls to TestCase methods show as warnings, but
    will be resolved at run-time.
    """
    def __init__(self, *args, **kwargs):
        args = (args[0][5:],)  # Strip the 'test_' prefix from the method name
        super().__init__(*args, **kwargs)
        self.test_name: str = ''  # TestName reference
        self.input_data: Optional[List[Matrix]] = None  # Regression input matrices
        self.regression_manager: Optional[RegressionManager] = None  # Regression manager
        self.subject: T = None  # Subject to test

    @TestRegression
    def default_setup(self):
        pass

    def before_each(self, test_info):
        self.input_data = self.get_input_data()
        self.subject = self.instantiate_subject()
        self.test_name = self._testMethodName.replace('()', '')
        self.regression_manager = RegressionManager(self.get_reference_dir(), self.test_name)

    def after_each(self, test_info):
        # Check if the test contains the regression tag
        method = getattr(self, self.test_name, None)
        if has_tag(method, REGRESSION_TAG):
            self.run_regression()

    @abstractmethod
    def setup_regressions(self, subject: T, input_data: List[Matrix]):
        """
        Setup the given regressions. First run the algorithm on the data. Then
        add all components of the model via add_regression or other add methods.

        :param subject:     Algorithm to run.
        :param input_data:  Input data.
        """
        pass

    @abstractmethod
    def get_datasets(self) -> List[TestDataset]:
        """
        Get the input datasets used for the algorithm tests.

        :return:    Paths to input data.
        """
        pass

    @abstractmethod
    def instantiate_subject(self) -> T:
        """
        Create an instance of the subject.
        :return:
        """
        pass

    def run_regression(self):
        """
        Run the set-up algorithm.
        """
        try:
            self.setup_regressions(self.subject, self.input_data)
        except:
            print_stack_trace()
            self.fail('Setting up regression group failed.')

        self.regression_manager.run_assertions()

    def get_input_data(self):
        """
        Get the input data.

        :return:    Input data.
        """
        return [dataset.load() for dataset in self.get_datasets()]

    def get_reference_dir(self) -> str:
        """
        Get reference file directory.
        Constructed based on the package name and removes "Test" from each class name.

        :return:    Reference file directory.
        """
        # Get the fully-qualified name of the class (in dotted form)
        fully_qualified_name = self.__class__.__module__ + '.' + self.__class__.__qualname__

        # Rename the test directory to wai.ma to match the source class structure
        fully_qualified_name = fully_qualified_name.replace("test.", "wai.ma.")

        # Split into sub-level parts
        name_parts = fully_qualified_name.split('.')

        # Remove any "private" levels (module/package names starting with _)
        name_parts = [part for part in name_parts if not part.startswith('_')]

        # Remove the "Test" suffix from the class name
        if name_parts[-1].endswith('Test'):
            name_parts[-1] = name_parts[-1][:-4]

        return os.path.join('resources', 'regression', *name_parts)

    def add_regression(self, tag: str, value: Union[Matrix, real]):
        """
        Add a new Matrix/real with a tag to the regression manager.

        :param tag:     Tag for the value.
        :param value:   Value to check with the reference.
        """
        self.regression_manager.add(tag, value)

    def setUp(self) -> None:
        self.before_each(None)

    def tearDown(self) -> None:
        self.after_each(None)



#  _YGradientGLSWSavitzkyGolayFilter.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from ..core import real
from ..core.error import MatrixAlgorithmsError
from ..core.matrix import Matrix, factory
from ._AbstractTransformation import AbstractTransformation


class YGradientGLSWSavitzkyGolayFilter(AbstractTransformation):
    def __init__(self):
        super().__init__()
        self.coef: List[real] = [real(2.0 / 10.0),
                                 real(1.0 / 10.0),
                                 real(0.0),
                                 real(-1.0 / 10.0),
                                 real(-2.0 / 10.0)]

    def configure(self, data: Matrix):
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        mat_extended = self.extend_matrix(data)
        result = factory.zeros_like(mat_extended)

        for i in range(2, mat_extended.num_rows() - 2):
            row_smoothed = self.smooth_row(i, mat_extended)
            result.set_row(i, row_smoothed)

        orig_sized_matrix = self.shrink_matrix(result)
        return orig_sized_matrix

    def shrink_matrix(self, result: Matrix) -> Matrix:
        return result.get_sub_matrix((2, result.num_rows() - 2), (0, result.num_columns()))

    def extend_matrix(self, data: Matrix) -> Matrix:
        first_row: Matrix = data.get_row(0)
        last_row: Matrix = data.get_row(data.num_rows() - 1)
        return first_row\
            .concat_along_rows(first_row)\
            .concat_along_rows(data)\
            .concat_along_rows(last_row)\
            .concat_along_rows(last_row)

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        raise MatrixAlgorithmsError('Inverse transformation of Savitzky Golay is not available.')

    def smooth_row(self, i: int, matrix: Matrix) -> Matrix:
        res: Matrix = factory.zeros(1, matrix.num_columns())

        window_size: int = len(self.coef)
        for m in range(window_size):
            coef_idx = window_size - 1 - m
            row_idx = i - (m - 2)
            row = matrix.get_row(row_idx)
            row_scaled = row.mul(self.coef[coef_idx])
            res = res.add(row_scaled)

        return res



#  _YGradientGLSWTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, TypeVar

from ._GLSWTest import GLSWTest
from ...test.misc import TestDataset
from wai.ma.algorithm.glsw import YGradientGLSW
from wai.ma.core.matrix import Matrix

T = TypeVar('T', bound=YGradientGLSW)


class YGradientGLSWTest(GLSWTest[T]):
    def setup_regressions(self, subject: YGradientGLSW, input_data: List[Matrix]):
        # Get inputs: Simulate second instrument as x1 with noise
        X: Matrix = input_data[0]
        y: Matrix = input_data[1]

        # Init GLSW
        subject.initialize(X, y)

        # Add regressions
        self.add_glsw_regressions(subject, X)

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS, TestDataset.BOLTS_RESPONSE]

    def instantiate_subject(self) -> YGradientGLSW:
        return YGradientGLSW()


#  _NIPALS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from enum import Enum, auto
from typing import Optional, List

from ._AbstractMultiResponsePLS import AbstractMultiResponsePLS
from ...core import ZERO, real, ONE
from ...core.matrix import Matrix, factory
from ...transformation import Standardize


class NIPALS(AbstractMultiResponsePLS):
    """
    Nonlinear Iterative Partial Least Squares

    Implementation oriented at scikit-learn's NIPALS implementation:
    <a href="https://github.com/scikit-learn/scikit-learn/blob/ed5e127b/sklearn/cross_decomposition/pls_.py#L455">Github scikit-learn NIPALS</a>

    Parameters:
    - tol: Iterative convergence tolerance
    - maxIter: Maximum number of iterations
    - normYWeights: Flat to normalize Y weights
    - deflationMode: Mode for Y matrix deflation. Can be either CANONICAL or REGRESSION
    """
    def __init__(self):
        super().__init__()
        self.X_scores: Optional[Matrix] = None  # Scores on X
        self.Y_scores: Optional[Matrix] = None  # Scores on Y
        self.X_loadings: Optional[Matrix] = None  # Loadings on X
        self.Y_loadings: Optional[Matrix] = None  # Loadings on Y
        self.X_weights: Optional[Matrix] = None  # Weights on X
        self.Y_weights: Optional[Matrix] = None  # Weights on Y
        self.X_rotations: Optional[Matrix] = None  # Projection of X into latent space
        self.Y_rotations: Optional[Matrix] = None  # Projection of Y into latent space
        self.X: Optional[Matrix] = None  # Training points
        self.coef: Optional[Matrix] = None  # Regression coefficients
        self.tol: real = real(1e-6)  # Inner NIPALS loop improvement tolerance
        self.max_iter: int = 500  # Inner NIPALS loop maximum number of iterations
        self.norm_Y_weights: bool = False  # Flag to normalize Y weights
        self.standardize_X: Optional[Standardize] = None  # Standarize X tranformation
        self.standardize_Y: Optional[Standardize] = None  # Standardize Y transformation
        self.deflation_mode: DeflationMode = DeflationMode.REGRESSION

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.tol = real(1e-6)
            self.max_iter = 500
            self.norm_Y_weights = False
            self.standardize_X = Standardize()
            self.standardize_Y = Standardize()
            self.deflation_mode = DeflationMode.REGRESSION
        else:
            return super().initialize(predictors, response)

    @staticmethod
    def validate_max_iter(value: int) -> bool:
        return value >= 0

    @staticmethod
    def validate_tol(value: real) -> bool:
        return value >= ZERO

    def get_min_columns_response(self) -> int:
        return 1

    def get_max_columns_response(self) -> int:
        return -1

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        # Init
        X: Matrix = predictors
        X = self.standardize_X.transform(X)
        Y: Matrix = response
        Y = self.standardize_Y.transform(Y)

        # Dimensions
        num_rows: int = X.num_rows()
        num_features: int = X.num_columns()
        num_classes: int = Y.num_columns()
        num_components: int = self.num_components

        # Init matrices
        self.X_scores = factory.zeros(num_rows, num_components)  # T
        self.Y_scores = factory.zeros(num_rows, num_components)  # U

        self.X_weights = factory.zeros(num_features, num_components)  # W
        self.Y_weights = factory.zeros(num_classes, num_components)  # C

        self.X_loadings = factory.zeros(num_features, num_components)  # P
        self.Y_loadings = factory.zeros(num_classes, num_components)  # Q

        yk_loading: Matrix = factory.zeros(num_classes, 1)

        eps: real = real(1e-10)
        for k in range(num_components):
            if Y.transpose().mul(Y).all(lambda e: e < eps):
                self.logger.warning('Y residual constant at iteration ' + str(k))
                break

            res: NipalsLoopResult = self.nipals_loop(X, Y)
            xk_weight: Matrix = res.X_weights
            yk_weight: Matrix = res.Y_weights

            # Calculate latent X and Y scores
            xk_score: Matrix = X.mul(xk_weight)
            yk_score: Matrix = Y.mul(yk_weight).div(yk_weight.norm2_squared())

            if xk_score.norm2_squared() < eps:
                self.logger.warning('X scores are null at component ' + str(k))
                break

            # Deflate X
            xk_loading: Matrix = X.transpose().mul(xk_score).div(xk_score.norm2_squared())
            X = X.sub(xk_score.mul(xk_loading.transpose()))

            # Deflate Y
            if self.deflation_mode is DeflationMode.CANONICAL:
                yk_loading: Matrix = Y.transpose().mul(yk_score).div(yk_score.norm2_squared())
                Y = Y.sub(yk_score.mul(yk_loading.transpose()))
            elif self.deflation_mode is DeflationMode.REGRESSION:
                yk_loading: Matrix = Y.transpose().mul(xk_score).div(xk_score.norm2_squared())
                Y = Y.sub(xk_score.mul(yk_loading.transpose()))

            # Store results
            self.X_scores.set_column(k, xk_score)
            self.Y_scores.set_column(k, yk_score)
            self.X_weights.set_column(k, xk_weight)
            self.Y_weights.set_column(k, yk_weight)
            self.X_loadings.set_column(k, xk_loading)
            self.Y_loadings.set_column(k, yk_loading)

        self.X = X
        self.X_rotations = self.X_weights.mul((self.X_loadings.transpose().mul(self.X_weights)).pseudo_inverse())
        if Y.num_columns() > 1:
            self.Y_rotations = self.Y_weights.mul((self.Y_loadings.transpose().mul(self.Y_weights)).pseudo_inverse())
        else:
            self.Y_rotations = factory.filled(1, 1, ONE)

        # Calculate regression coefficients
        y_stds: Matrix = self.standardize_Y.get_std_devs()
        self.coef = self.X_rotations.mul(self.Y_loadings.transpose()).scale_by_row_vector(y_stds)
        return None

    def nipals_loop(self, X: Matrix, Y: Matrix) -> 'NipalsLoopResult':
        """
        Perform the inner NIPALS loop.

        :param X:   Predictors matrix.
        :param Y:   Response matrix.
        :return:    NipalsLoopResult.
        """
        iterations: int = 0

        y_score: Matrix = Y.get_column(0)  # (y scores)
        X_weight_old: Matrix = factory.zeros(X.num_columns(), 1)
        X_p_inv: Optional[Matrix] = None
        Y_p_inv: Optional[Matrix] = None

        eps: real = real(1e-16)

        # Repeat 1) - 3) until convergence: either change of u is lower than m_Tol or maximum
        # number of iterations has been reached (m_MaxIter)
        while True:
            # 1) Update X weights
            if self.get_weight_calculation_mode() is WeightCalculationMode.CCA:
                if X_p_inv is None:
                    # sklearn uses pinv here which ojAlgo implicitly does
                    X_p_inv = X.inverse()
                X_weight: Matrix = X_p_inv.mul(y_score)
            else:  # PLS
                X_weight: Matrix = X.transpose().mul(y_score).div(y_score.norm2_squared())

            # Add eps if necessary to converge to a more acceptable solution
            if X_weight.norm2_squared() < eps:
                X_weight = X_weight.add(eps)

            # Normalize
            X_weight = X_weight.div(X_weight.norm2() + eps)

            # 2) Calculate latent X scores
            X_score: Matrix = X.mul(X_weight)

            # 3) Update Y weights
            if self.get_weight_calculation_mode() is WeightCalculationMode.CCA:
                if Y_p_inv is None:
                    # sklearn uses pinv here which ojAlgo implicitly does
                    Y_p_inv = Y.inverse()
                Y_weight: Matrix = Y_p_inv.mul(X_score)
            else:  # PLS
                # WeightCalculationMode A: Regress each Y column on xscore
                Y_weight: Matrix = Y.transpose().mul(X_score).div(X_score.norm2_squared())

            # Normalise Y weights
            if self.norm_Y_weights:
                Y_weight = Y_weight.div(Y_weight.norm2() + eps)

            # 4) Calculate ykScores
            Y_score: Matrix = Y.mul(Y_weight).div(Y_weight.norm2_squared() + eps)

            X_weight_diff: Matrix = X_weight.sub(X_weight_old)

            if X_weight_diff.norm2_squared() < self.tol or Y.num_columns() == 1:
                break

            if iterations >= self.max_iter:
                break

            # Update stopping conditions
            X_weight_old = X_weight
            iterations += 1

        return NipalsLoopResult(X_weight, Y_weight, iterations)

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        X: Matrix = self.standardize_X.transform(predictors)

        Y_means: Matrix = self.standardize_Y.get_means()
        Y_hat: Matrix = X.mul(self.coef).add_by_vector(Y_means)
        return Y_hat

    def do_transform(self, predictors: Matrix) -> Matrix:
        X: Matrix = self.standardize_X.transform(predictors)

        # Apply rotations
        X_scores: Matrix = X.mul(self.X_rotations)
        return X_scores

    def do_transform_response(self, response: Matrix) -> Matrix:
        Y: Matrix = self.standardize_Y.transform(response)

        # Apply rotations
        Y_scores: Matrix = Y.mul(self.Y_rotations)
        return Y_scores

    def get_matrix_names(self) -> List[str]:
        return ['T', 'U', 'P', 'Q']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        if name == 'T':
            return self.X_scores
        elif name == 'U':
            return self.Y_scores
        elif name == 'P':
            return self.X_loadings
        elif name == 'Q':
            return self.Y_loadings
        return None

    def has_loadings(self) -> bool:
        return True

    def reset(self):
        super().reset()
        self.X_scores = None
        self.Y_scores = None
        self.X_loadings = None
        self.Y_loadings = None
        self.X_weights = None
        self.Y_weights = None
        self.coef = None
        self.X = None
        self.X_rotations = None
        self.Y_rotations = None
        self.standardize_X = Standardize()
        self.standardize_Y = Standardize()

    def get_loadings(self) -> Optional[Matrix]:
        return self.X_loadings

    def can_predict(self) -> bool:
        return True

    def get_weight_calculation_mode(self) -> 'WeightCalculationMode':
        return WeightCalculationMode.PLS  # Mode A in sklearn


class NipalsLoopResult:
    """
    NIPALS loop result: x and y weight matrices and number of iterations.
    """
    def __init__(self, X_weights: Matrix, Y_weights: Matrix, iterations: int):
        self.X_weights: Matrix = X_weights
        self.Y_weights: Matrix = Y_weights
        self.iterations: int = iterations


class DeflationMode(Enum):
    """
    Deflation mode enum.
    """
    CANONICAL = auto()
    REGRESSION = auto()


class WeightCalculationMode(Enum):
    """
    Mode for x/y-weight calculation
    """
    PLS = auto()
    CCA = auto()



#  factory.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Factory methods for Matrix objects.
"""

import builtins
from numbers import Number
from typing import List, Union

import numpy as np
from numpy import random

from ._Matrix import Matrix
from .._types import real


def create(data: Union[np.ndarray, List[List[Number]]]) -> Matrix:
    """
    Create a matrix from a given set of data.

    :param data:    An ndarray or a 2D list of numbers.
    :return:        A matrix.
    """
    return Matrix(np.array(data, real, copy=True))


def eye(rows: int, columns: int = -1) -> Matrix:
    """
    Creates an identity matrix (1s on the diagonal
    and 0s everywhere else.

    :param rows:        The number of rows in the matrix.
    :param columns:     The number of columns in the matrix.
                        Default is equal to rows.
    :return:            The matrix.
    """
    if columns == -1:
        columns = rows
    return Matrix(np.eye(rows, columns, dtype=real))


def eye_like(other: Matrix) -> Matrix:
    """
    Creates an identity matrix the same shape as another matrix.

    :param other:   The matrix whose dimensions to copy.
    :return:        The matrix.
    """
    return eye(other.num_rows(), other.num_columns())


from_raw = create


def zeros(rows: int, columns: int = -1) -> Matrix:
    """
    Creates a matrix filled with 0s.

    :param rows:        The number of rows the matrix should have.
    :param columns:     The number of columns the matrix should have.
                        Defaults to the number of rows.
    :return:            The matrix.
    """
    if columns == -1:
        columns = rows
    return create([[0 for _ in builtins.range(columns)]
                   for _ in builtins.range(rows)])


def zeros_like(other: Matrix) -> Matrix:
    """
    Creates a matrix filled with 0s, the same shape as the
    given matrix.

    :param other:   The matrix whose shape should be copied.
    :return:        The matrix.
    """
    return zeros(other.num_rows(), other.num_columns())


def filled(rows: int, columns: int, initial_value: Number) -> Matrix:
    """
    Creates a matrix filled with an initial value.

    :param rows:            The number of rows.
    :param columns:         The number of columns.
    :param initial_value:   The initial value to fill the matrix with.
    :return:                The matrix.
    """
    return Matrix(np.full((rows, columns), real(initial_value), real))


def filled_like(other: Matrix, initial_value: Number) -> Matrix:
    """
    Creates a matrix the same shape as another, but filled with
    an initial value.

    :param other:           The matrix whose shape should be copied.
    :param initial_value:   The value to fill the matrix with.
    :return:                The matrix.
    """
    return filled(other.num_rows(), other.num_columns(), initial_value)


def from_row(vector: List[Number]) -> Matrix:
    """
    Creates a row vector matrix from a list of values.

    :param vector:  The values to put in the matrix.
    :return:        The row vector matrix.
    """
    return create([vector])


def from_column(vector: List[Number]) -> Matrix:
    """
    Creates a column vector matrix from a list of values.

    :param vector:  The values to put in the matrix.
    :return:        The column vector matrix.
    """
    return create([[v] for v in vector])


def randn(rows: int, columns: int, seed: int = 1, mean: real = real(0), std: real = real(1)) -> Matrix:
    """
    Creates a matrix filled with random values taken from a normal distribution.

    :param rows:        The number of rows.
    :param columns:     The number of columns.
    :param seed:        The seed value to initialise the PRNG with.
    :param mean:        The mean of the normal distribution.
    :param std:         The standard deviation of the normal distribution.
    :return:            The matrix.
    """
    random.seed(seed)
    data = random.rand(rows, columns)
    np.multiply(data, std)
    np.add(data, mean)
    return Matrix(data)


def randn_like(other: Matrix, seed: int, mean: real = real(0), std: real = real(1)) -> Matrix:
    """
    Creates a matrix filled with random values taken from a normal distribution,
    in the same shape as another matrix.

    :param other:       The matrix whose shape should be copied.
    :param seed:        The seed value to initialise the PRNG with.
    :param mean:        The mean of the normal distribution.
    :param std:         The standard deviation of the normal distribution.
    :return:            The matrix.
    """
    return randn(other.num_rows(), other.num_columns(), seed, mean, std)


def rand(rows: int, columns: int, seed: int = 1) -> Matrix:
    """
    Creates a matrix filled with random values taken from a uniform distribution.

    :param rows:        The number of rows.
    :param columns:     The number of columns.
    :param seed:        The seed value to initialise the PRNG with.
    :return:            The matrix.
    """
    random.seed(seed)
    return Matrix(random.rand(rows, columns))


def rand_like(other: Matrix, seed: int) -> Matrix:
    """
    Creates a matrix filled with random values taken from a uniform distribution.

    :param other:       The matrix whose shape should be copied.
    :param seed:        The seed value to initialise the PRNG with.
    :return:            The matrix.
    """
    return rand(other.num_rows(), other.num_columns(), seed)


def diag(vector: Matrix) -> Matrix:
    """
    Creates a diagonal matrix from a column vector.

    :param vector:
    :return:
    """
    n = vector.num_rows()
    result = zeros(n, n)
    for i in builtins.range(n):
        result.set(i, i, vector.get(i, 0))
    return result


def range(rows: int, columns: int, start: int) -> Matrix:
    """
    Creates a matrix where each element is the next number in a
    sequence starting with the given value. Increments along the
    columns before wrapping to the next row.

    :param rows:        The number of rows.
    :param columns:     The number of columns.
    :param start:       The value of the first element.
    :return:            The matrix.
    """
    return create([[x for x in builtins.range(s, s + columns)]
                   for s in builtins.range(start, start + rows * columns, columns)])


def create_spectrum(wave_numbers: List[Number], amplitudes: List[Number]) -> Matrix:
    """
    Creates a spectrum matrix, which has two columns (the wave-numbers and the
    amplitudes). The provided wave-number and amplitude lists must be the same
    length.

    :param wave_numbers:    The wave-numbers of the spectrum.
    :param amplitudes:      The amplitudes of the waves.
    :return:                The spectrum matrix.
    """
    if len(wave_numbers) != len(amplitudes):
        raise ValueError('Must have equal number of amplitudes and wave-numbers')

    return Matrix(np.array([[w, a] for w, a in zip(wave_numbers, amplitudes)], dtype=real))



#  _SavitzkyGolay.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List

import numpy as np

from ..core import real
from ..core.matrix import factory, helper, Matrix
from ._AbstractTransformation import AbstractTransformation


class SavitzkyGolay(AbstractTransformation):
    def __init__(self):
        super().__init__()
        self.polynomial_order: int = 2
        self.derivative_order: int = 1
        self.num_points_left: int = 3
        self.num_points_right: int = 3
        self.coefficients: Optional[Matrix] = None

    def __setattr__(self, key, value):
        # Validate values
        validator_name = 'validate_' + key
        if hasattr(self, validator_name):
            validator = getattr(self, validator_name)
            validator(value)

            # If validation passes, reset
            self.reset()

        super().__setattr__(key, value)

    @staticmethod
    def validate_polynomial_order(value: int):
        if value < 2:
            raise ValueError('polynomial_order must be at least 2')

    @staticmethod
    def validate_derivative_order(value: int):
        if value < 0:
            raise ValueError('derivative_order must be at least 0')

    @staticmethod
    def validate_num_points_left(value: int):
        if value < 0:
            raise ValueError('num_points_left must be at least 0')

    @staticmethod
    def validate_num_points_right(value: int):
        if value < 0:
            raise ValueError('num_points_right must be at least 0')

    def reset(self):
        super().reset()
        self.coefficients = None

    def configure(self, data: Matrix):
        if self.coefficients is None:
            self.coefficients = factory.create([determine_coefficients(self.num_points_left,
                                                                       self.num_points_right,
                                                                       self.polynomial_order,
                                                                       self.derivative_order)])
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        smoothed_columns = []

        window_width = self.coefficients.num_columns()
        num_output_columns = data.num_columns() - window_width + 1
        for i in range(num_output_columns):
            column = data.get_sub_matrix((0, data.num_rows()), (i, i + window_width))
            column.mul_by_vector_modify(self.coefficients)
            column = column.sum(1)
            smoothed_columns.append(column)

        return helper.multi_concat(1, *smoothed_columns)

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        pass


def determine_coefficients(num_left: int, num_right: int, poly_order: int, der_order: int) -> List[real]:
    result: List[real] = [real(0)] * (num_left + num_right + 1)

    if len(result) == 1:
        result[0] = real(1)
        return result

    A_dim = poly_order + 1
    A = factory.zeros(A_dim, A_dim)
    for i in range(A_dim):
        for j in range(A_dim):
            sum = real(0)
            for k in range(-num_left, num_right + 1):
                sum += np.power(k, i + j)
            A.set(i, j, sum)

    b = factory.zeros(A_dim, 1)
    b.set(der_order, 0, real(1))

    solution = helper.solve(A, b)

    for i in range(-num_left, num_right + 1):
        sum = real(0)
        for j in range(A_dim):
            sum += solution.get(j, 0) * np.power(i, j)
        result[i + num_left] = sum

    return result



#  _MatrixRegression.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple, List

from wai.ma.core.matrix import Matrix, helper
from ._AbstractRegression import AbstractRegression


class MatrixRegression(AbstractRegression):
    """
    Matrix regression implementation.
    """
    def __init__(self, path: str, actual: Matrix):
        super().__init__(path, actual)

    def check_equals(self, expected: Matrix, actual: Matrix):
        # Check for shape
        if not expected.same_shape_as(actual):
            self.fail('Shapes of the expected and actual matrices do not match.\n'
                      + 'Expected shape: ' + expected.shape_string() +'\n'
                      + 'Actual shape: ' + actual.shape_string())

        # Check for values
        diff: Matrix = expected.sub(actual).abs()
        which = diff.which(lambda v: v > AbstractRegression.EPSILON)
        if len(which) > 0:
            self.fail('Regression ' + self.get_path() + ' failed.\n'
                      + 'Absolute differences:\n'
                      + self.indices_to_string(which, diff))

    def read_expected(self, path: str):
        return helper.read(path, False, ',')

    def write_expected(self, path: str, expected):
        helper.write(expected, path, False, ',', 10, False)

    def get_filename_extension(self) -> str:
        return 'csv'

    def indices_to_string(self, which: List[Tuple[int, int]], diff: Matrix) -> str:
        string = '(<row>, <column>): <absolute difference>\n'
        string += '\n'.join(('({i}, {j}): {diff}'.format_map({'i': i, 'j': j, 'diff': diff.get(i, j)})
                            for i, j in which))
        return string



#  _DoubleRegression.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from wai.ma.core import real
from ._AbstractRegression import AbstractRegression


class RealRegression(AbstractRegression):
    """
    Real value regression implementation.
    """
    def __init__(self, path: str, actual: real):
        super().__init__(path, actual)

    def check_equals(self, expected, actual):
        self.assertAlmostEqual(expected, actual, delta=AbstractRegression.EPSILON)

    def read_expected(self, path: str):
        with open(path, 'r') as file:
            return real(file.readline())

    def write_expected(self, path: str, expected):
        with open(path, 'w') as file:
            file.writelines(str(expected))

    def get_filename_extension(self) -> str:
        return 'txt'



#  utils.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Helper functions.
"""
from numbers import Number
from typing import List, NoReturn, Tuple, Optional
import numpy as np

from ._types import real

NAN = 'NaN'
NEGATIVE_INFINITY = '-Infinity'
POSITIVE_INFINITY = '+Infinity'


def initial_index(size: int) -> List[int]:
    """
    Initial index, filled with values from 0 to size - 1.
    """
    return [i for i in range(size)]


def sort_left_right_and_center(array: List[real], index: List[int], l: int, r: int) -> int:
    """
    Sorts left, right and center elements only, returns resulting center as pivot.
    """
    c: int = (l + r) // 2
    conditional_swap(array, index, l, c)
    conditional_swap(array, index, l, r)
    conditional_swap(array, index, c, r)
    return c


def swap(index: List[int], l: int, r: int) -> NoReturn:
    """
    Swaps two elements in the given integer array.
    """
    help: int = index[l]
    index[l] = index[r]
    index[r] = help


def conditional_swap(array: List[real], index: List[int], left: int, right: int) -> NoReturn:
    """
    Conditional swap for quick-sort.
    """
    if array[index[left]] > array[index[right]]:
        swap(index, left, right)


def partition(array, index: List[int], l: int, r: int, pivot: Optional[real] = None) -> int:
    """
    Partitions the instances around a pivot. Used by quicksort and kthSmallestValue.

    :param array:   The list of reals to be sorted.
    :param index:   The index into the list of reals.
    :param l:       The first index of the subset.
    :param r:       The last index of the subset.

    :return:        The index of the middle element.
    """
    if isinstance(array[0], real):
        return _partition_real(array, index, l, r, pivot)
    else:
        return _partition_int(array, index, l, r)


def _partition_real(array: List[real], index: List[int], l: int, r: int, pivot: real) -> int:
    """
    Partitions the instances around a pivot. Used by quicksort and kthSmallestValue.

    :param array:   The list of reals to be sorted.
    :param index:   The index into the list of reals.
    :param l:       The first index of the subset.
    :param r:       The last index of the subset.

    :return:        The index of the middle element.
    """
    r -= 1
    while True:
        l += 1
        while array[index[l]] < pivot:
            l += 1
        r -= 1
        while array[index[r]] > pivot:
            r -= 1
        if l >= r:
            return l
        swap(index, l, r)


def _partition_int(array: List[int], index: List[int], l: int, r: int) -> int:
    """
    Partitions the instances around a pivot. Used by quicksort and kthSmallestValue.

    :param array:   The array of integers to be sorted.
    :param index:   The index into the array of integers.
    :param l:       The first index of the subset.
    :param r:       The last index of the subset.

    :return:        The index of the middle element.
    """
    pivot: real = real(array[index[(l + r) // 2]])

    while l < r:
        while array[index[l]] < pivot and l < r:
            l += 1
        while array[index[r]] > pivot and l < r:
            r -= 1
        if l < r:
            swap(index, l, r)
            l += 1
            r -= 1
    if l == r and array[index[r]] > pivot:
        r -= 1

    return r


def quick_sort(array: List[real], index: List[int], left: int, right: int):
    """
    Implements quicksort with median-of-three method and explicit sort for problems
    of size three or less.

    :param array:   The list of reals to be sorted.
    :param index:   The index into the list of reals.
    :param left:    The first index of the subset to be sorted.
    :param right:   The last index of the subset to be sorted.
    :return:
    """
    diff: int = right - left

    if diff == 0:
        # No need to do anything
        return
    elif diff == 1:
        # Swap two elements if necessary
        conditional_swap(array, index, left, right)
        return
    elif diff == 2:
        # Just need to sort three elements
        conditional_swap(array, index, left, left + 1)
        conditional_swap(array, index, left, right)
        conditional_swap(array, index, left + 1, right)
        return
    else:
        # Establish pivot
        pivot_location: int = sort_left_right_and_center(array, index, left, right)

        # Move pivot to right, partition, and restore pivot
        swap(index, pivot_location, right - 1)
        center: int = partition(array, index, left, right, array[index[right - 1]])
        swap(index, center, right - 1)

        # Sort recursively
        quick_sort(array, index, left, center - 1)
        quick_sort(array, index, center + 1, right)


def sort(array: List[real]) -> List[int]:
    """
    Sorts a given array of doubles in ascending order and returns an array of
    integers with the positions of the elements of the original array in the
    sorted array. NOTE THESE CHANGES: the sort is no longer stable and it
    doesn't use safe floating-point comparisons anymore. Occurrences of
    Double.NaN are treated as Double.MAX_VALUE.

    :param array:   This list is not changed by the method!
    :return:        A list of integers with the positions in the sorted list.
    """
    index = initial_index(len(array))
    if len(array) > 1:
        array = list(array)
        quick_sort(array, index, 0, len(array) - 1)
    return index


def linear_regression(x: List[real], y: List[real]) -> Tuple[real, real]:
    """
    Calculates the slope and intercept between the two lists.

    :param x:   The first list, representing the X values.
    :param y:   The second list, representing the Y values.
    :return:    Intercept, slope
    """
    n = len(x)
    x_times_y = [x * y for x, y in zip(x, y)]

    a = (sum(y) * sum_of_squares(x) - sum(x) * sum(x_times_y))\
        / (n * sum_of_squares(x) - pow(sum(x), 2))

    b = (n * sum(x_times_y) - sum(x) * sum(y))\
        / (n * sum_of_squares(x) - pow(sum(x), 2))

    return a, b


def sum_of_squares(l: List[real]) -> real:
    """
    Returns the sum of the squares of all the elements in the list.

    :param l:   The list to work on.
    :return:    The sum.
    """
    return sum((x * x for x in l), real(0))


def to_real_array(array: List[Number]) -> List[real]:
    """
    Turns a Number list into a list of reals.

    :param array:   The list to convert.
    :return:        The converted list.
    """
    return [real(x) for x in array]


to_real_list = to_real_array


def max_index(reals: List[real]) -> int:
    """
    Returns index of maximum element in a given list of reals. First maximum
    is returned.

    :param reals:   The list of reals.
    :return:        The index of the maximum element.
    """
    maximum: real = real(0)
    max_index = 0

    for i in range(len(reals)):
        if i == 0 or reals[i] > maximum:
            max_index = 1
            maximum = reals[i]

    return max_index


def real_to_string_fixed(value: real, after_decimal_point: int) -> str:
    """
    Rounds a real and converts it into a string. Always displays the specified
    number of decimals.

    :param value:                   The real value.
    :param after_decimal_point:     The number of digits permitted after the decimal
                                    point; if -1 then all decimals are displayed; also
                                    if number > Long.MAX_VALUE. TODO: Revisit Long.MAX_VALUE comment.
    :return:                        The real as a formatted string.
    """
    # Special numbers
    if value == real(NAN):
        return NAN
    elif value == real(NEGATIVE_INFINITY):
        return NEGATIVE_INFINITY
    elif value == real(POSITIVE_INFINITY):
        return POSITIVE_INFINITY

    if after_decimal_point < 0:
        return str(value)

    return np.format_float_positional(value,
                                      precision=after_decimal_point,
                                      unique=False,
                                      fractional=True,
                                      trim='k')


def get_list_dimensions(array: List[any]) -> int:
    """
    Returns the dimensions of the given list.

    :param array:   The list to determine dimensions for.
    :return:        The dimensions of the list.
    """
    if isinstance(array[0], list):
        return 1 + get_list_dimensions(array[0])
    else:
        return 1


def list_to_string(array: List[any], output_class: bool = False) -> str:
    """
    Returns the given list in a string representation.

    :param array:           The list to return in a string representation.
    :param output_class:    Whether to output the class name instead of calling
                            the object's 'toString()' method. TODO: Revisit toString() comment.
    :return:                The list as a string.
    """
    # TODO
    raise NotImplementedError


def exp(x: real) -> real:
    """
    Returns Euler's number raised to the given power a.

    :param x:   The exponent.
    :return:    The result of e^x.
    """
    return np.exp(x)


def sqrt(x: real) -> real:
    """
    Returns the square-root of the given number.

    :param x:   The number to find the root of.
    :return:    The square root of x.
    """
    return np.sqrt(x)



#  _AbstractPLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from typing import TypeVar, List, Optional

from ...test import AbstractRegressionTest
from ...test.misc import Test, Tags, TestDataset
from wai.ma.algorithm.pls import AbstractPLS
from wai.ma.core.matrix import Matrix

T = TypeVar('T', bound=AbstractPLS)


class AbstractPLSTest(AbstractRegressionTest[T]):
    @Test
    def check_transformed_num_components(self):
        X: Matrix = self.input_data[0]
        Y: Matrix = self.input_data[1]

        for i in range(1, 5):
            self.subject.num_components = i
            self.subject.initialize(X, Y)
            transform: Matrix = self.subject.transform(X)
            self.assertEqual(i, transform.num_columns())

            # Reset
            self.subject = self.instantiate_subject()

    @staticmethod
    def merge_if_not_empty(tag_1: str, tag_2: str) -> str:
        if tag_1 != '' and tag_2 != '':
            return os.sep.join((tag_1, tag_2))
        else:
            return tag_1 + tag_2

    def setup_regressions(self, subject: AbstractPLS, input_data: List[Matrix]):
        # Extract data
        X: Matrix = input_data[0]
        y: Matrix = input_data[1]

        # Initialise PLS
        results: Optional[str] = subject.initialize(X, y)
        if results is not None:
            self.fail('Algorithm#initialize failed with result: ' + results)

        self.add_default_pls_matrices(subject, X)

    def add_default_pls_matrices(self, algorithm: AbstractPLS, x: Matrix, sub_tag: str = ''):
        """
        Adds default PLS matrices, that is predictions, transformations, loadings and
        model parameter matrices.
        """
        self.add_predictions(algorithm, x, sub_tag)
        self.add_transformation(algorithm, x, sub_tag)
        self.add_loadings(algorithm, sub_tag)
        self.add_matrices(algorithm, sub_tag)

    def add_transformation(self, algorithm: AbstractPLS, x: Matrix, sub_tag: str = ''):
        """
        Add transformation to the regression group.
        """
        self.add_regression(self.merge_if_not_empty(sub_tag, Tags.TRANSFORM), algorithm.transform(x))

    def add_predictions(self, algorithm: AbstractPLS, x: Matrix, sub_tag: str = ''):
        """
        Add predictions to the regression group.
        """
        # Add predictions
        if algorithm.can_predict():
            preds: Matrix = algorithm.predict(x)
            self.add_regression(self.merge_if_not_empty(sub_tag, Tags.PREDICTIONS), preds)

    def add_loadings(self, algorithm: AbstractPLS, sub_tag: str = ''):
        """
        Add loadings to the regression group.
        """
        # Add loadings
        if algorithm.has_loadings():
            self.add_regression(self.merge_if_not_empty(sub_tag, Tags.LOADINGS), algorithm.get_loadings())

    def add_matrices(self, algorithm: AbstractPLS, sub_tag: str = ''):
        """
        Add model matrices to the regression group.
        """
        # Add matrices
        for matrix_name in algorithm.get_matrix_names():
            tag: str = Tags.MATRIX + '-' + matrix_name
            self.add_regression(self.merge_if_not_empty(sub_tag, tag), algorithm.get_matrix(matrix_name))

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS, TestDataset.BOLTS_RESPONSE]



#  _CCAFilterTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from ..test import AbstractRegressionTest, T
from ..test.misc import TestRegression, Tags, TestDataset
from wai.ma.algorithm import CCAFilter
from wai.ma.core.matrix import Matrix


class CCAFilterTest(AbstractRegressionTest[CCAFilter]):
    @TestRegression
    def lambda_X_10(self):
        self.subject.lambda_X = 10

    @TestRegression
    def lambda_Y_10(self):
        self.subject.lambda_Y = 10

    def setup_regressions(self, subject: T, input_data: List[Matrix]):
        X: Matrix = self.input_data[0]
        Y: Matrix = self.input_data[1]

        self.subject.initialize(X, Y)

        self.add_regression(Tags.TRANSFORM, self.subject.transform(X))
        self.add_regression(Tags.PROJECTION + '-X', self.subject.proj_X)
        self.add_regression(Tags.PROJECTION + '-Y', self.subject.proj_Y)

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS, TestDataset.BOLTS_RESPONSE]

    def instantiate_subject(self) -> CCAFilter:
        return CCAFilter()



#  _PolyKernel.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ...core import ZERO, real, NAN, ONE
from ...core.matrix import Matrix
from ._AbstractKernel import AbstractKernel


class PolyKernel(AbstractKernel):
    """
    Poly Kernel.

    K(x,y)=(gamma*x^T*y + coef_0)^d
    """
    def __init__(self):
        self.degree: int = 3
        self.coef_0: real = ZERO
        self.gamma: real = NAN
        super().__init__()

    def __setattr__(self, key, value):
        if key == 'degree' and value <= 0:
            return
        super().__setattr__(key, value)

    def apply_matrix(self, X: Matrix, Y: Optional[Matrix] = None) -> Matrix:
        if Y is None:
            Y = X
        if self.gamma is NAN:
            self.gamma = ONE / X.num_columns()
        result: Matrix = X.mul(Y.transpose())
        result = result.mul(self.gamma)
        result = result.add(self.coef_0)
        result = result.pow_elementwise(self.degree)
        return result

    def apply_vector(self, x: Matrix, y: Matrix) -> real:
        linear_term: real = x.vector_dot(y)
        if self.gamma is NAN:
            self.gamma = ONE / x.num_columns()
        return pow(self.gamma * linear_term + self.coef_0, self.degree)

    def __str__(self):
        return 'Polynomial Kernel: K(x,y)=(gamma*x^T*y + coef_0)^d, ' +\
               'gamma=' + str(self.gamma) + ', d=' + str(self.degree) + ', coef_0=' + str(self.coef_0)



#  _RBFKernel.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ...core import ONE, real, NAN
from ...core.matrix import Matrix
from ...core.utils import exp
from ._AbstractKernel import AbstractKernel


class RBFKernel(AbstractKernel):
    """
    Radial Basis Function Kernel.

    K(x,y) = exp(-1*||x - y||^2/(2*sigma^2))
    or
    K(x,y) = exp(-1*gamma*||x - y||^2), with gamma=2*sigma^2
    """
    def __init__(self):
        self.gamma: real = ONE  # Gamma parameter
        super().__init__()

    def apply_vector(self, x: Matrix, y: Matrix) -> real:
        norm_2: real = x.sub(y).norm2()
        if self.gamma is NAN:
            self.gamma = ONE / x.num_columns()
        return exp(-ONE * self.gamma * norm_2)

    def __str__(self):
        return 'RBF Kernel: K(x,y) = exp(-1*gamma*||x - y||^2), gamma=' + str(self.gamma)


from ..core import Filter
from ..core.matrix import Matrix


class Downsample(Filter):
    """
    Filter which gets every Nth row from a matrix, starting at a given index.
    """
    def __init__(self):
        self.start_index: int = 0  # The index to start sampling from
        self.step: int = 1  # The step-size between samples

    def transform(self, predictors: Matrix) -> Matrix:
        rows = [i for i in range(self.start_index, predictors.num_rows(), self.step)]

        return predictors.get_rows(rows)



#  _OPLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ._AbstractPLSTest import AbstractPLSTest
from ...test.misc import TestRegression
from wai.ma.algorithm.pls import OPLS, NIPALS, KernelPLS, PLS1, SIMPLS, SparsePLS, PRM, CCARegression


class OPLSTest(AbstractPLSTest[OPLS]):
    def check_transformed_num_components(self):
        # Do nothing since OPLS.transform(X) removes the signal from X_test that is
        # orthogonal to y_train and does not change its shape
        pass

    @TestRegression
    def base_NIPALS(self):
        self.subject.base_PLS = NIPALS()

    @TestRegression
    def base_KernelPLS(self):
        self.subject.base_PLS = KernelPLS()

    @TestRegression
    def base_PLS1(self):
        self.subject.base_PLS = PLS1()

    @TestRegression
    def base_SIMPLS(self):
        self.subject.base_PLS = SIMPLS()

    @TestRegression
    def base_CCARegression(self):
        self.subject.base_PLS = CCARegression()

    @TestRegression
    def base_PRM(self):
        self.subject.base_PLS = PRM()

    @TestRegression
    def base_SparsePLS(self):
        self.subject.base_PLS = SparsePLS()

    def instantiate_subject(self) -> OPLS:
        return OPLS()



#  _FastICATest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from ...test import AbstractRegressionTest
from ...test.misc import TestRegression, TestDataset
from wai.ma.algorithm.ica import FastICA, Algorithm
from wai.ma.algorithm.ica.approxfun import LogCosH, Cube, Exponential
from wai.ma.core.matrix import Matrix


class FastICATest(AbstractRegressionTest[FastICA]):
    @TestRegression
    def deflation(self):
        self.subject.algorithm = Algorithm.DEFLATION

    @TestRegression
    def parallel(self):
        self.subject.algorithm = Algorithm.PARALLEL

    @TestRegression
    def white_false(self):
        self.subject.whiten = False

    @TestRegression
    def logcosh(self):
        self.subject.fun = LogCosH()

    @TestRegression
    def cube(self):
        self.subject.fun = Cube()

    @TestRegression
    def exp(self):
        self.subject.fun = Exponential()

    def setup_regressions(self, subject: FastICA, input_data: List[Matrix]):
        X: Matrix = input_data[0]

        transform: Matrix = self.subject.transform(X)
        self.add_regression('transform', transform)
        self.add_regression('components', self.subject.components)
        self.add_regression('mixing', self.subject.mixing)
        self.add_regression('sources', self.subject.sources)

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS]

    def instantiate_subject(self) -> FastICA:
        return FastICA()



#  _InvalidShapeError.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


class InvalidShapeError(RuntimeError):
    def __init__(self, message: str, *matrix_shapes: str):
        if matrix_shapes is None or len(matrix_shapes) == 0:
            super().__init__('Invalid shape ' + message)
        else:
            super().__init__('Invalid shapes '
                             + ', '.join([m for m in matrix_shapes])
                             + ' '
                             + message
                             )



#  _TestDataset.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from enum import Enum
from unittest import TestCase

from wai.ma.meta import print_stack_trace
from wai.ma.core.matrix import Matrix, helper


class TestDataset(Enum):
    """
    Test datasets that represent a file in the 'resources' folder.
    """
    BOLTS = os.path.join('resources', 'bolts.csv')
    BOLTS_RESPONSE = os.path.join('resources', 'bolts_response.csv')

    def __str__(self):
        return self.value

    def load(self) -> Matrix:
        """
        Load a matrix from a given input path.

        :return:    Matrix stored in input path.
        """
        try:
            return helper.read(self.value, True, ',')
        except:
            print_stack_trace()
            TestCase.fail(None, 'Failed to load dataset at path: ' + self.value)



#  _AbstractRegression.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from abc import abstractmethod, ABCMeta
from unittest import TestCase

from wai.ma.meta import print_stack_trace
from wai.ma.core import real


class AbstractRegression(TestCase, metaclass=ABCMeta):
    """
    Class representing an abstract regression test regarding a single reference file.
    """

    EPSILON: real = real(1e-6)  # Tolerance

    def __init__(self, path: str, actual):
        super().__init__()
        self.path: str = path + '.' + self.get_filename_extension()  # Regression path
        self.actual = actual  # Actual result
        self.expected = None  # Expected result
        self.__load_expected()

    @staticmethod
    def ensure_dir_exists(path: str):
        sub_path: str = os.path.dirname(path)
        os.makedirs(sub_path, exist_ok=True)

    def get_path(self) -> str:
        return self.path

    @abstractmethod
    def check_equals(self, expected, actual):
        """
        Check the actual and expected results for equality.
        """
        pass

    def run_assertions(self):
        """
        Run all assertions.
        """
        self.check_equals(self.expected, self.actual)

    @abstractmethod
    def read_expected(self, path: str):
        """
        Read the expected result from the reference path.

        :param path:    Reference file path.
        :return:        Expected/Reference result.
        """
        pass

    @abstractmethod
    def write_expected(self, path: str, expected):
        """
        Write the expected result in the case when no expected result could be found.

        :param path:        Reference file path.
        :param expected:    Expected object.
        """
        pass

    def __load_expected(self):
        """
        Loads the expected results.
        """
        try:
            self.expected = self.read_expected(self.path)
        except FileNotFoundError:
            print('File <' + self.path + '> does not exist yet. Creating new reference.')
            self.create_new_reference(self.path)
            self.expected = self.actual
        except:
            print_stack_trace()
            self.fail('Failed to load reference file: ' + self.path)

    def create_new_reference(self, path: str):
        """
        Create new reference file.

        :param path:    File path to store the reference at.
        """
        try:
            self.ensure_dir_exists(path)
            self.write_expected(path, self.actual)
        except:
            print_stack_trace()
            self.fail('Failed to create new reference file: ' + path)

    @abstractmethod
    def get_filename_extension(self) -> str:
        """
        Gets the filename extension. E.g. one of [txt, csv, dat, ref, ...]

        :return:    Filename extension
        """
        pass


#  _switch.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import threading
from enum import Enum
from typing import Optional, List, Dict

context_stacks: Dict[int, List['SwitchContextManager']] = dict()


def add_to_stack(context_manager: 'SwitchContextManager'):
    # Get the thread we're running on
    current_thread: int = threading.get_ident()

    # Ensure there is a context stack for that thread
    if current_thread not in context_stacks:
        context_stacks[current_thread] = []

    # Put ourselves on the stack
    context_stacks[current_thread].append(context_manager)


def pop_from_stack():
    # Get the thread we're running on
    current_thread: int = threading.get_ident()

    # Pop the top of the stack
    context_stacks[current_thread].pop()


def get_current_context_manager() -> 'SwitchContextManager':
    # Get the thread we're running on
    current_thread: int = threading.get_ident()

    # Peek the top of the stack
    return context_stacks[current_thread][-1]


def switch(on) -> 'SwitchContextManager':
    """
    Sugar method for setting up the switch-case context manager.
    Should be called as:

        with switch(on):
            if case([cases]):
                ...
                break_()
            if case([more cases]):
                ...
                break_()
            if default():
                ...

    :param on:  The value to be switched on.
    :return:    The switch context manager.
    """
    return SwitchContextManager(on)


class SwitchContextManager:
    """
    Context manager for switch-case blocks.
    """
    def __init__(self, on: Enum):
        self.on = on
        self.matched = False

    def __enter__(self):
        # Add ourselves as the current context manager for the thread
        add_to_stack(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Remove ourselves as the current context manager for the thread
        pop_from_stack()

        # If a break statement was used to exit,
        # don't propagate the exception
        return exc_type is BreakSwitchException

    def __call__(self, *cases: Optional[Enum]) -> bool:
        # If we match any of the provided cases, always match from
        # now on. No cases provided is the default case.
        if len(cases) == 0 or any(case is self.on for case in cases):
            self.matched = True

        # Allows for fall-through
        return self.matched


def case(*cases: Optional[Enum]) -> bool:
    # Get the current context
    current_context: SwitchContextManager = get_current_context_manager()

    # Call it to check the cases provided
    return current_context(*cases)


def default():
    return case()


def break_():
    """
    Breaks from the switch block.
    """
    raise BreakSwitchException


class BreakSwitchException(Exception):
    pass



#  _FastICA.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from enum import Enum, auto
from typing import Optional, Tuple

from math import sqrt

from ...core import Filter
from .._AbstractAlgorithm import AbstractAlgorithm
from .approxfun import NegEntropyApproximationFunction, LogCosH
from ...core.error import MatrixAlgorithmsError
from ...core.matrix import Matrix
from ...core.matrix.factory import randn, zeros
from ...transformation import Center


class FastICA(AbstractAlgorithm, Filter):
    def __init__(self):
        super().__init__()
        self.num_components: int = 5
        self.whiten: bool = True
        self.fun: Optional[NegEntropyApproximationFunction] = None
        self.max_iter: int = 500
        self.tol: float = 1e-4
        self.components: Optional[Matrix] = None
        self.sources: Optional[Matrix] = None
        self.algorithm: Optional[Algorithm] = None
        self.center: Optional[Center] = None
        self.whitening: Optional[Matrix] = None
        self.mixing: Optional[Matrix] = None

    @staticmethod
    def validate_num_components(value: int) -> bool:
        return value >= 1

    @staticmethod
    def validate_whiten(value: bool) -> bool:
        return True

    @staticmethod
    def validate_max_iter(value: int) -> bool:
        return value >= 0

    @staticmethod
    def validate_tol(value: float) -> bool:
        return value >= 0.0

    def initialize(self):
        super().initialize()
        self.num_components = 5
        self.max_iter = 500
        self.fun = LogCosH()
        self.tol = 1e-4
        self.whiten = True
        self.algorithm = Algorithm.DEFLATION
        self.center = Center()

    def reset(self):
        super().reset()
        self.components = None
        self.mixing = None
        self.whitening = None

    def configure(self, X: Matrix):
        X = X.transpose()

        n: int = X.num_rows()
        p: int = X.num_columns()
        unmixing: Optional[Matrix] = None
        X1: Optional[Matrix] = None

        min_NP: int = min(n, p)

        if not self.whiten:
            self.num_components = min_NP
            self.logger.warning('Ignoring num_components when $whiten=False')

        if self.num_components > min_NP:
            self.logger.warning('num_components is too large and will be set to ' + str(min_NP))
            self.num_components = min_NP

        # Whiten data
        if self.whiten:
            X = self.center.transform(X.transpose()).transpose()
            U: Matrix = X.svd_U()
            d: Matrix = X.get_singular_values()
            k: int = min_NP  # Rank k
            d = d.get_rows((0, k))  # Only get non-zero singular values
            d_inv_elements: Matrix = d.apply_elementwise(lambda a: 1.0 / a)
            tmp: Matrix = U.scale_by_row_vector(d_inv_elements).transpose()
            self.whitening = tmp.get_rows((0, min(tmp.num_rows(), self.num_components)))

            X1 = self.whitening.mul(X)
            X1 = X1.mul(sqrt(p))
        else:
            X1 = X

        # Randomly initialise weights from normal dist
        W_init: Matrix = randn(self.num_components, self.num_components, 1)

        # Use deflation algorithm
        if self.algorithm is Algorithm.DEFLATION:
            unmixing = self.deflation(X1, W_init)
        elif self.algorithm is Algorithm.PARALLEL:  # Use parallel algorithm
            unmixing = self.parallel(X1, W_init)

        # Compute sources and components
        if self.whiten:
            self.sources = unmixing.mul(self.whitening).mul(X).transpose()
            self.components = unmixing.mul(self.whitening)
        else:
            self.sources = unmixing.mul(X).transpose()
            self.components = unmixing

        self.mixing = self.components.inverse()
        self.initialised = True

    def deflation(self, X: Matrix, W_init: Matrix) -> Matrix:
        """
        Deflationary FastICA.

        :param X:       Input.
        :return:        Weights
        """
        W: Matrix = zeros(self.num_components, self.num_components)

        for j in range(self.num_components):
            w: Matrix = W_init.get_row(j).transpose().copy()
            w = w.div(w.pow_elementwise(2).sum(-1).sqrt().as_real())
            for i in range(self.max_iter):
                res: Tuple[Matrix, Matrix] = self.fun.apply(w.transpose().mul(X).transpose())

                gwtx: Matrix = res[0]
                g_wtx: Matrix = res[1]

                w1: Matrix = X.scale_by_row_vector(gwtx).mean(1).sub(w.mul(g_wtx.mean()))
                w1 = self.decorrelate(w1, W, j)

                w1 = w1.div(w1.pow_elementwise(2).sum(-1).sqrt().as_real())
                lim = w1.mul_elementwise(w).sum(-1).abs().sub(1.0).abs().as_real()

                w = w1
                if lim < self.tol:
                    break

            W.set_row(j, w)

        return W

    def parallel(self, X: Matrix, W_init: Matrix) -> Matrix:
        """
        Parallel FastICA.

        :param X:       Input.
        :param W_init:  Initial weight matrix.
        :return:        Weight.
        """
        W: Matrix = self.symmetric_decorrelation(W_init)

        p: int = X.num_columns()

        for i in range(self.max_iter):
            res: Tuple[Matrix, Matrix] = self.fun.apply(W.transpose().mul(X))
            gwtx: Matrix = res[0]
            g_wtx: Matrix = res[1]

            arg: Matrix = gwtx.mul(X.transpose()).div(p).sub(W.scale_by_column_vector(g_wtx))  # Scale by row?
            W1: Matrix = self.symmetric_decorrelation(arg)
            lim = W1.mul(W.transpose()).diag().abs().sub(1.0).abs().max()
            W = W1
            if lim < self.tol:
                break

        return W

    def decorrelate(self, w: Matrix, W: Matrix, j: int) -> Matrix:
        """
        Orthonormalise w wrt the first j columns of W.

        :param w:   w vector.
        :param W:   W matrix.
        :param j:   First j columns.
        :return:    Orthonormalised w.
        """
        if j == 0:
            return w
        Wp: Matrix = W.get_rows((0, j))

        if j == 1:
            s: Matrix = w.transpose().mul(Wp.transpose()).mul(Wp).transpose()
            return w.sub(s)

        sub: Matrix = w.transpose().mul(Wp.transpose()).mul(Wp).transpose()
        return w.sub(sub)

    def symmetric_decorrelation(self, W: Matrix) -> Matrix:
        """
        W = (W * W.T)^(-1/2)

        :param W:   Weight matrix.
        :return:    Decorrelated weight matrix.
        """
        wwt: Matrix = W.mul(W.transpose())
        eigvals: Matrix = wwt.get_eigenvalues_sorted_ascending()
        eigvecs: Matrix = wwt.get_eigenvectors_sorted_ascending()
        s: Matrix = eigvals
        u: Matrix = eigvecs

        # np.dot(np.dot(u * (1. / np.sqrt(s)), u.T), W)
        s_sqrt: Matrix = s.sqrt()
        s_inv: Matrix = s_sqrt.apply_elementwise(lambda x: 1.0 / x)
        u_mule_s: Matrix = u.scale_by_row_vector(s_inv)
        return u_mule_s.mul(u.transpose()).mul(W)

    def to_string(self) -> str:
        return 'FastICA{' +\
               'num_components=' + str(self.num_components) +\
               ', whiten=' + str(self.whiten) +\
               ', fun=' + str(self.fun) +\
               ', max_iter=' + str(self.max_iter) +\
               ', tol=' + str(self.tol) +\
               '}'

    def do_transform(self, data: Matrix) -> Matrix:
        """
        Transform a matrix.

        :param data:    The original data to transform.
        :return:        The transformed data.
        """
        return self.sources

    def transform(self, data: Matrix) -> Matrix:
        self.reset()
        self.configure(data)
        result: Matrix = self.do_transform(data)

        return result

    def reconstruct(self) -> Matrix:
        if self.is_initialised():
            return self.center.inverse_transform(self.sources.mul(self.mixing.transpose()).transpose()).transpose()
        else:
            raise MatrixAlgorithmsError('FastICA has not yet been initialized!')


class Algorithm(Enum):
    PARALLEL = auto(),
    DEFLATION = auto()


#  _AbstractTransformationTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from ..test import AbstractRegressionTest, T
from ..test.misc import Test, Tags, TestDataset
from wai.ma.core.matrix import Matrix


class AbstractTransformationTest(AbstractRegressionTest[T]):
    """
    Abstract tranformation test. Regression for transform and inverse-transform.
    """
    @Test
    def check_inv_transform_eq_input(self):
        input: Matrix = self.input_data[0]
        transform: Matrix = self.subject.transform(input)
        inverse_transform: Matrix = self.subject.inverse_transform(transform)

        # Check if input == inverse_transform
        is_equal: bool = input.sub(inverse_transform).abs().all(lambda v: v < 1e-7)
        self.assertTrue(is_equal)

    def setup_regressions(self, subject: T, input_data: List[Matrix]):
        X: Matrix = input_data[0]
        subject.configure(X)

        transform: Matrix = subject.transform(X)
        inverse_transform: Matrix = subject.inverse_transform(transform)

        self.add_regression(Tags.TRANSFORM, transform)
        self.add_regression(Tags.INVERSE_TRANSFORM, inverse_transform)

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS]


#  _GLSWTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import TypeVar, List

from ...test import AbstractRegressionTest
from ...test.misc import Tags, TestRegression, TestDataset
from wai.ma.algorithm.glsw import GLSW
from wai.ma.core.matrix import Matrix, factory

T = TypeVar('T', bound=GLSW)


class GLSWTest(AbstractRegressionTest[T]):
    def add_glsw_regressions(self, subject: GLSW, x: Matrix):
        # Add regressions
        self.add_regression(Tags.TRANSFORM, subject.transform(x))
        self.add_regression(Tags.PROJECTION, subject.G)

    @TestRegression
    def alpha_1(self):
        self.subject.alpha = 1

    @TestRegression
    def alpha_100(self):
        self.subject.alpha = 100

    def setup_regressions(self, subject: GLSW, input_data: List[Matrix]):
        # Get inputs: Simulate second instrument as x1 with noise
        x_first_instrument: Matrix = input_data[0]
        x_second_instrument: Matrix = x_first_instrument.add(factory.randn_like(x_first_instrument, 0, 0.0, 0.1))

        # Init GLSW
        subject.initialize(x_first_instrument, x_second_instrument)

        self.add_glsw_regressions(subject, x_first_instrument)

    def get_datasets(self) -> List[TestDataset]:
        return [TestDataset.BOLTS]

    def instantiate_subject(self) -> GLSW:
        return GLSW()


#  _LoggingObject.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABC

from ..meta import JavaObject
from .java import Logger


class LoggingObject(JavaObject):
    """
    Ancestor of objects with logging support.
    """
    def __init__(self):
        super().__init__()
        self.logger: Logger = Logger.get_logger(self.__class__.__name__)  # The logger to use.
        self.debug: bool = False  # Whether to output debug information.

    def __construct__(self):
        super().__construct__()
        self.initialize()
        self.reset()

    def __setattr__(self, key, value):
        # Validate values
        validator_name = 'validate_' + key
        if hasattr(self, validator_name):
            validator = getattr(self, validator_name)
            if not validator(value):
                raise ValueError('Validation of ' + key + ' failed')

            # If validation passes, reset, unless in __init__ phase
            if self.init_complete():
                self.reset()

        super().__setattr__(key, value)

    def initialize(self):
        """
        For initialising members.

        Default implementation does nothing.
        """
        pass

    def reset(self):
        """
        For resetting data structures when changing parameters.

        Default implementation does nothing.\
        """
        pass



#  _OPLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, List

from ._AbstractSingleResponsePLS import AbstractSingleResponsePLS
from ._AbstractPLS import AbstractPLS
from ._PLS1 import PLS1
from ...core import real
from ...core.matrix import Matrix, factory
from ...meta import switch, default, case


class OPLS(AbstractSingleResponsePLS):
    """
    OPLS algorithm.

    See here:
    <a href="https://onlinelibrary.wiley.com/doi/pdf/10.1002/cem.695">
        Orthogonal Projections to latent structures (O-PLS)
    </a>
    """
    def __init__(self):
        super().__init__()
        self.P_orth: Optional[Matrix] = None  # The P matrix
        self.T_orth: Optional[Matrix] = None  # The T matrix
        self.W_orth: Optional[Matrix] = None  # The W matrix
        self.X_osc: Optional[Matrix] = None  # Data with orthogonal signal components removed
        self.base_PLS: Optional[AbstractPLS] = None  # Base PLS that is trained on the cleaned data

    @staticmethod
    def validate_base_PLS(value: AbstractPLS) -> bool:
        return True

    def reset(self):
        """
        Resets the member variables.
        """
        super().reset()

        self.P_orth = None
        self.W_orth = None
        self.T_orth = None

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.base_PLS = PLS1()
        else:
            super().initialize(predictors, response)

    def get_matrix_names(self) -> List[str]:
        """
        Returns all the available matrices.

        :return:    The names of the matrices.
        """
        return ['P_orth',
                'W_orth',
                'T_orth']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        """
        Returns the matrix with the specified name.

        :param name:    The name of the matrix.
        :return:        The matrix, None if not available.
        """
        with switch(name):
            if case('P_orth'):
                return self.P_orth
            if case('W_orth'):
                return self.W_orth
            if case('T_orth'):
                return self.T_orth
            if default():
                return None

    def has_loadings(self) -> bool:
        """
        Whether the algorithm supports return of loadings.

        :return:    True if supported.
        """
        return True

    def get_loadings(self) -> Optional[Matrix]:
        """
        Returns the loadings, if available.

        :return:    The loadings, None if not available.
        """
        return self.P_orth

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initialises using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        X: Matrix = predictors.copy()
        X_trans: Matrix = X.transpose()
        y: Matrix = response

        # Init
        self.W_orth = factory.zeros(predictors.num_columns(), self.num_components)
        self.P_orth = factory.zeros(predictors.num_columns(), self.num_components)
        self.T_orth = factory.zeros(predictors.num_rows(), self.num_components)

        w: Matrix = X_trans.mul(y).mul(self.inv_L2_squared(y)).normalized()

        for current_component in range(self.num_components):
            # Calculate scores vector
            t: Matrix = X.mul(w).mul(self.inv_L2_squared(w))

            # Calculate loadings of X
            p: Matrix = X_trans.mul(t).mul(self.inv_L2_squared(t))

            # Orthogonalise weight
            w_orth: Matrix = p.sub(w.mul(w.transpose().mul(p).mul(self.inv_L2_squared(w)).as_real()))
            w_orth = w_orth.normalized()
            t_orth: Matrix = X.mul(w_orth).mul(self.inv_L2_squared(w_orth))
            p_orth: Matrix = X_trans.mul(t_orth).mul(self.inv_L2_squared(t_orth))

            # Remove orthogonal components from X
            X = X.sub(t_orth.mul(p_orth.transpose()))
            X_trans = X.transpose()

            # Store results
            self.W_orth.set_column(current_component, w_orth)
            self.T_orth.set_column(current_component, t_orth)
            self.P_orth.set_column(current_component, p_orth)

        self.X_osc = X.copy()
        self.base_PLS.initialize(self.do_transform(predictors), response)

        return None

    def inv_L2_squared(self, v: Matrix) -> real:
        """
        Get the inverse of the squared L2 norm.

        :param v:   Input vector.
        :return:    1.0 / norm2(v)^2.
        """
        l2: real = v.norm2_squared()
        return real(1 / l2)

    def do_transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the data.

        :param predictors:  The input data.
        :return:            The transformed data and the prediction.
        """
        # Remove signal from X_test that is orthogonal to y_train
        # X_clean = X_test - X_test*W_orth*P_orth^T
        T: Matrix = predictors.mul(self.W_orth)
        X_orth: Matrix = T.mul(self.P_orth.transpose())
        return predictors.sub(X_orth)

    def can_predict(self) -> bool:
        """
        Returns whether the algorithm can make predictions.

        :return:    True if can make predictions.
        """
        return True

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        X_transformed: Matrix = self.transform(predictors)
        return self.base_PLS.predict(X_transformed)



#  _JavaObjectMeta.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABCMeta


class JavaObjectMeta(ABCMeta):
    """
    Meta-class for creating objects that are more Java-like.
    """
    def __call__(cls, *args, **kwargs):
        # Create the object as normal
        instance = super().__call__(*args, **kwargs)

        # If the class defines a constructor method, call it
        if hasattr(instance, '__construct__'):
            instance.__construct__(*args, **kwargs)

        # Return the new instance
        return instance



#  _SparsePLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, Set, List

from ._AbstractSingleResponsePLS import AbstractSingleResponsePLS
from ._NIPALS import NIPALS
from ...core import ZERO, real
from ...core.matrix import Matrix, factory
from ...meta import switch, case, default
from ...transformation import Standardize


class SparsePLS(AbstractSingleResponsePLS):
    """
    Sparse PLS algorithm.

    See here:
    <a href="https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2810828/">
        Sparse partial least squares regression for simultaneous dimension reduction and variable selection
    </a>

    Implementation was oriented at the R SPLS package, which implements the above
    mentioned paper:
    <a href="https://github.com/cran/spls">
        Sparse Partial Least Squares (SPLS) Regression and Classification
    </a>

    The lambda parameter controls the features sparseness. For sufficiently small
    lambda, all features will be selected and the algorithm results are equal
    to NIPALS'.
    """
    def __init__(self):
        super().__init__()
        self.B_pls: Optional[Matrix] = None
        self.tol: real = ZERO  # NIPALS tolerance threshold
        self.max_iter: int = 0  # NIPALS max iterations
        self.lambda_: real = ZERO  # Sparsity parameter. Determines sparseness
        self.A: Optional[Set[int]] = None
        self.W: Optional[Matrix] = None  # Loadings
        self.standardize_X: Optional[Standardize] = None  # Standardize X
        self.standardize_Y: Optional[Standardize] = None  # Standardize Y

    @staticmethod
    def validate_max_iter(value: int) -> bool:
        return value >= 0

    @staticmethod
    def validate_tol(value: real) -> bool:
        return value >= 0

    @staticmethod
    def validate_lambda_(value: real) -> bool:
        return value >= 0

    def reset(self):
        """
        Resets the member variables.
        """
        super().reset()
        self.B_pls = None
        self.A = None
        self.W = None
        self.standardize_X = Standardize()
        self.standardize_Y = Standardize()

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.lambda_ = real(0.5)
            self.tol = real(1e-7)
            self.max_iter = 500
            self.standardize_X = Standardize()
            self.standardize_Y = Standardize()
        else:
            return super().initialize(predictors, response)

    def get_matrix_names(self) -> List[str]:
        """
        Returns the names of all the available matrices.

        :return:    The names of the matrices.
        """
        return ['W',
                'B']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        """
        Returns the matrix with the specified name.

        :param name:    The name of the matrix.
        :return:        The matrix, None if not available.
        """
        with switch(name):
            if case('W'):
                return self.W
            if case('B'):
                return self.B_pls
            if default():
                return None

    def has_loadings(self) -> bool:
        """
        Whether the algorithm supports the return of loadings.

        :return:    True if supported.
        """
        return True

    def get_loadings(self) -> Optional[Matrix]:
        """
        Returns the loadings, if available.

        :return:    The loadings, None if not available.
        """
        return self.W

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initialises using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        X: Matrix = self.standardize_X.transform(predictors)
        y: Matrix = self.standardize_Y.transform(response)
        X_j: Matrix = X.copy()
        y_j: Matrix = y.copy()
        self.A = set()
        self.B_pls = factory.zeros(X.num_columns(), y.num_columns())
        self.W = factory.zeros(X.num_columns(), self.num_components)

        for k in range(self.num_components):
            w_k: Matrix = self.get_direction_vector(X_j, y_j, k)
            self.W.set_column(k, w_k)

            if self.debug:
                self.check_direction_vector(w_k)

            self.collect_indices(w_k)

            X_A: Matrix = self.get_column_submatrix_of(X)
            self.B_pls = factory.zeros(X.num_columns(), y.num_columns())
            B_pls_A: Matrix = self.get_regression_coefficient(X_A, y, k)

            # Fill self.B_pls values at non zero indices with estimated
            # regression coefficients
            idx_counter: int = 0
            for idx in sorted(self.A):
                self.B_pls.set_row(idx, B_pls_A.get_row(idx_counter))
                idx_counter += 1

            # Deflate
            y_j = y.sub(X.mul(self.B_pls))

        if self.debug:
            self.logger.info('Selected following features ' +\
                             '(' + str(len(self.A)) + '/' + str(X.num_columns()) + '): ')
            l: List[str] = [str(a) for a in sorted(self.A)]
            self.logger.info(','.join(l))

        return None

    def get_regression_coefficient(self, X_A: Matrix, y: Matrix, k: int) -> Matrix:
        """
        Calculate NIPALS regression coefficient.

        :param X_A:     Predictors subset.
        :param y:       Current response vector.
        :param k:       PLS iteration.
        :return:        B_pls (NIPALS regression coefficients)
        """
        num_components: int = min(X_A.num_columns(), k + 1)
        nipals: NIPALS = NIPALS()
        nipals.max_iter = self.max_iter
        nipals.tol = self.tol
        nipals.num_components = num_components
        nipals.initialize(X_A, y)
        return nipals.coef

    def get_column_submatrix_of(self, X: Matrix) -> Matrix:
        """
        Get the column submatrix of X given by the indices in self.A.

        :param X:   Input matrix.
        :return:    Submatrix of X.
        """
        X_A: Matrix = factory.zeros(X.num_rows(), len(self.A))
        col_count: int = 0
        for i in sorted(self.A):
            col: Matrix = X.get_column(i)
            X_A.set_column(col_count, col)
            col_count += 1
        return X_A

    def get_row_submatrix_of(self, X: Matrix) -> Matrix:
        """
        Get the row submatrix of X given by the indices in self.A.

        :param X:   Input matrix.
        :return:    Submatrix of X.
        """
        X_A: Matrix = factory.zeros(len(self.A), X.num_columns())
        row_count: int = 0
        for i in sorted(self.A):
            row: Matrix = X.get_row(i)
            X_A.set_row(row_count, row)
            row_count += 1
        return X_A

    def collect_indices(self, w: Matrix):
        """
        Collect indices based on the current non zero indices in w and self.B_pls.

        :param w:   Direction vector.
        """
        self.A.clear()
        self.A.update(w.where_vector(lambda d: abs(d) > 1e-6))
        self.A.update(self.B_pls.where_vector(lambda d: abs(d) > 1e-6))

    def check_direction_vector(self, w: Matrix):
        """
        Check if the direction vector fulfills w^Tw=1.

        :param w:   Direction vector.
        """
        # Test if w^Tw = 1
        if w.norm2_squared() - 1 > 1e-6:
            self.logger.warning("Direction vector condition w'w=1 was violated.")

    def get_direction_vector(self, X: Matrix, y_j: Matrix, k: int) -> Matrix:
        """
        Compute the direction vector.

        :param X:       Predictors.
        :param y_j:     Current deflated response.
        :param k:       Iteration.
        :return:        Direction vector.
        """
        Z_p: Matrix = X.transpose().mul(y_j)
        z_norm: real = Z_p.abs().median()  # R package spls uses median norm
        Z_p = Z_p.div(z_norm)
        Z_P_sign: Matrix = Z_p.sign()
        val_b: Matrix = Z_p.abs().sub(self.lambda_ * Z_p.abs().max())

        # Collect indices where val_b is >= 0
        idxs: List[int] = val_b.where_vector(lambda d: d >= 0)
        pre_mul: Matrix = val_b.mul_elementwise(Z_P_sign)
        c: Matrix = factory.zeros(Z_p.num_rows(), 1)
        for idx in idxs:
            val: real = pre_mul.get(idx, 0)
            c.set(idx, 0, val)

        return c.div(c.norm2_squared())  # Rescale c and use as estimated direction vector

    def do_transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        num_components: int = self.num_components
        T: Matrix = factory.zeros(predictors.num_rows(), num_components)
        X: Matrix = predictors.copy()
        for k in range(num_components):
            w_k: Matrix = self.W.get_column(k)
            t_k: Matrix = X.mul(w_k)
            T.set_column(k, t_k)

            p_k: Matrix = X.transpose().mul(t_k).div(t_k.norm2_squared())
            X = X.sub(t_k.mul(p_k.transpose()))

        return T

    def can_predict(self) -> bool:
        """
        Returns whether the algorithm can make predictions.

        :return:    True if can make predictions.
        """
        return True

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        X: Matrix = self.standardize_X.transform(predictors)

        X_A: Matrix = self.get_column_submatrix_of(X)
        B_A: Matrix = self.get_row_submatrix_of(self.B_pls)

        y_means: Matrix = self.standardize_Y.get_means()
        y_std: Matrix = self.standardize_Y.get_std_devs()
        y_hat: Matrix = X_A.mul(B_A).scale_by_row_vector(y_std).add_by_vector(y_means)

        return y_hat



#  _RegressionManager.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from typing import List, Union

from .regression import AbstractRegression, MatrixRegression, RealRegression
from wai.ma.core import real
from wai.ma.core.matrix import Matrix


class RegressionManager:
    """
    Group of regressions. Contains all parameters/outputs/model-matrices that
    need to be checked for the regression tests of a single model configuration.

    Each object that should be checked has to be added via one of the add methods.
    """
    def __init__(self, reference_dir: str, test_name: str):
        self.reference_dir: str = reference_dir
        self.regressions: List[AbstractRegression] = []
        self.test_name: str = test_name

    def add(self, tag: str, value: Union[Matrix, real]):
        path: str = self.construct_path(tag)
        if isinstance(value, Matrix):
            regression: AbstractRegression = MatrixRegression(path, value)
        else:
            regression: AbstractRegression = RealRegression(path, value)
        self.regressions.append(regression)

    def construct_path(self, tag: str) -> str:
        return os.path.join(self.reference_dir, self.test_name, tag)

    def run_assertions(self):
        for reg in self.regressions:
            reg.run_assertions()
